"""
Copyright © 2021 Jeff Kletsky. All Rights Reserved.

License for this software, part of the pyDE1 package, is granted under
GNU General Public License v3.0 only
SPDX-License-Identifier: GPL-3.0-only

Common events for the DE1 itself
"""

from copy import copy
from typing import Optional, List

from pyDE1.event_manager import EventPayload
from pyDE1.de1.c_api import API_MachineStates, API_Substates


class StateUpdate (EventPayload):
    """
    Derived from StateInfo

    Track previous centrally as it seems needed by so many consumers
    Tracking "at the source" should also be more robust

    No previous information is represented as NoRequest,NoState
    """
    def __init__(self, arrival_time: float,
                 state: API_MachineStates,
                 substate: API_Substates,
                 previous_state: Optional[API_MachineStates]=None,
                 previous_substate: Optional[API_Substates]=None,
                 ):
        super(StateUpdate, self).__init__(arrival_time=arrival_time)
        self._version = "1.0.0"  # Major version incremented on breaking change
        self.state = state
        self.substate = substate
        if previous_state is None:
            previous_state = API_MachineStates.NoRequest
        if previous_substate is None:
            previous_substate = API_Substates.NoState
        self.previous_state = previous_state
        self.previous_substate = previous_substate
        # External consumers don't have API_Substates.is_error
        self.is_error_state = (
                self.state == API_MachineStates.FatalError
                or self.substate.is_error
        )


class ShotSampleUpdate (EventPayload):
    """
    Derived from ShotSample

    See ShotSampleWithVolumeUpdates for API-visible class
    """
    _internal_only = True

    def __init__(self, arrival_time: float,
                 sample_time: int,
                 group_pressure: float,
                 group_flow: float,
                 mix_temp: float,
                 set_mix_temp: float,
                 set_head_temp: float,
                 set_group_pressure: float,
                 set_group_flow: float,
                 frame_number: int,
                 steam_temp: float,
                 ):
        super(ShotSampleUpdate, self).__init__(arrival_time=arrival_time)
        # TODO: This will need to be managed carefully as impacts subclass
        self._version = "1.0.0"  # Major version incremented on breaking change
        self.sample_time = sample_time
        self.group_pressure = group_pressure
        self.group_flow = group_flow
        self.mix_temp = mix_temp
        self.set_mix_temp = set_mix_temp
        self.set_head_temp = set_head_temp
        self.set_group_pressure = set_group_pressure
        self.set_group_flow = set_group_flow
        self.frame_number = frame_number
        self.steam_temp = steam_temp


class WaterLevelUpdate (EventPayload):
    """
    Derived from WaterLevels
    """
    def __init__(self, arrival_time:float,
                 level: float, start_fill_level:float):
        super(WaterLevelUpdate, self).__init__(arrival_time)
        self._version = "1.0.0"  # Major version incremented on breaking change
        self.level = level
        self.start_fill_level = start_fill_level


class ShotSampleWithVolumesUpdate (ShotSampleUpdate):
    """
    Delivered after ShotSampleUpdate,
    includes calculated then tracked volumes

    de1_time is preferred as it may eventually be time-base adjusted
    """
    _internal_only = False

    def __init__(self, shot_sample_update: ShotSampleUpdate,
                 volume_preinfuse: float,
                 volume_pour: float,
                 volume_total: float,
                 volume_by_frame: List[float],
                 ):
        # TODO: Is there a better way to do this?
        super(ShotSampleWithVolumesUpdate, self).__init__(
            arrival_time=shot_sample_update.arrival_time,
            sample_time=shot_sample_update.sample_time,
            group_pressure=shot_sample_update.group_pressure,
            group_flow=shot_sample_update.group_flow,
            mix_temp=shot_sample_update.mix_temp,
            set_mix_temp=shot_sample_update.set_mix_temp,
            set_head_temp=shot_sample_update.set_head_temp,
            set_group_pressure=shot_sample_update.set_group_pressure,
            set_group_flow=shot_sample_update.set_group_flow,
            frame_number=shot_sample_update.frame_number,
            steam_temp=shot_sample_update.steam_temp,
        )
        # TODO: This will need to be managed carefully as depends on super
        self._version = "1.1.0"  # Major version incremented on breaking change

        self.de1_time = self.arrival_time
        self.volume_preinfuse = volume_preinfuse
        self.volume_pour = volume_pour
        self.volume_total = volume_total
        self.volume_by_frames = copy(volume_by_frame)



"""
Copyright © 2021 Jeff Kletsky. All Rights Reserved.

License for this software, part of the pyDE1 package, is granted under
GNU General Public License v3.0 only
SPDX-License-Identifier: GPL-3.0-only
"""

import logging
from struct import unpack

import zlib
import binascii

# typedef struct {
#   U32 CheckSum;    // The checksum of the rest of the encrypted image. Includes "CheckSums" + "Data" fields, not "Header"
#   U32 BoardMarker; // 0xDE100001
#   U32 Version;     // The version of this image
#   U32 ByteCount;   // Number of bytes in image body, ignoring padding.
#   U32 CPUBytes;    // The first CPUBytes of the image are for the CPU. Remainder is for BLE.
#   U32 Unused;      // Blank spot for future extension. Always zero for now
#   U32 DCSum;       // Checksum of decrypted image
#   U8  IV[32];       // Initialization vector for the firmware
#   U32 HSum;        // Checksum of this header.
# } T_FirmwareHeader;


class FirmwareFile():

    def __init__(self, filename=None):
        self._filename = None
        self._checksum = None
        self._board_marker = None
        self._version = None
        self._byte_count = None
        self._cpu_bytes = None
        self._unused = None
        self._dc_sum = None
        self._initialization_vector = None
        self._header_checksum = None
        self._file_contents = None

        if filename is not None:
            self.filename = filename

    def _clear(self):
        self._checksum = None
        self._board_marker = None
        self._version = None
        self._byte_count = None
        self._cpu_bytes = None
        self._unknown = None
        self._dc_sum = None
        self._initialization_vector = None
        self._header_checksum = None

    @property
    def filename(self):
        return self._filename

    @filename.setter
    def filename(self, value):
        self._clear()
        self._filename = value

    @property
    def file_contents(self):
        if self._file_contents is None:
            self._load_from_file()
        return self._file_contents

    def _load_from_file(self):
        with open(self._filename, 'rb') as fh:
            # Header is
            #   7, 32-bit words (28 bytes)
            #   32-byte initialization vector
            #   1, 32-bit word of checksum (4 bytes)
            self._file_contents  = fh.read()
            print(f"File length: {len(self._file_contents)}")
            header = self._file_contents[0:64]
            (
                self._checksum,
                self._board_marker,
                self._version,
                self._byte_count,
                self._cpu_bytes,
                self._unknown,
                self._dc_sum,
                self._initialization_vector,
                self._header_checksum,
            ) = unpack('IIIIIII32sI', header)
            self._header = header
            # Not clear why
            print(f"binascii: {binascii.crc32(header[0:64]):08x}")
            print(f"zlib:     {zlib.crc32(header[0:64]):08x}")
            print(f"reported: {self._header_checksum:08x}")
            remainder = self._file_contents[64:]
            self._bytes_following = len(remainder)
            remainder_checksum = binascii.crc32(remainder)
            print()
            print(f"bcount: {self._byte_count}")
            print(f"length: {len(remainder)}")
            print(f"evaluated: {remainder_checksum:08x}")
            print(f"reported : {self._checksum:08x}")
            debug = True


if __name__ == '__main__':
    ff = FirmwareFile('../../bootfwupdate.dat')
    ff._load_from_file()

    print("done")

"""
Firmware update process:

11:38:29.743
Validate that the image is good, prior to writing (not being done now)

(what gets the DE1 into FW-update mode?? Is there any such thing?? GUI only??>)

11:38:54.329
Reboot DE1

11:39:38.806
Sleep the DE1
(Send fan temp of 60)

11:39:38.965
Send erase firmware: WI: 0, Erase: 1, Map: 1, FE: 0

11:39:48.828 - 12:30:02.530
Send firmware as write MMR 0x10 0x00 0000, 0010, 0020, ...
First packet is start of file itself (the header)

12:30:02.645
Enable map-request notifications

12:30:02.710
Use FirstError to determine if there are any errors
Iterate through them and "fix" them as needed

12:38:52.890
Reboot DE1
"""


"""
Copyright © 2021 Jeff Kletsky. All Rights Reserved.

License for this software, part of the pyDE1 package, is granted under
GNU General Public License v3.0 only
SPDX-License-Identifier: GPL-3.0-only

Provide a mapping from internal objects and properties, sufficient for both
creation of outgoing dict, as well as validation of incoming requests
(Resource, key, target object, property name, value type, min, max, post_execute)

Provide "mock" classes for use in inbound API process

NB: Must be imported AFTER any imports of DE1, Scale, FlowSequencer, ...
"""
import logging
import multiprocessing, multiprocessing.connection
import sys
from typing import Optional, Union, NamedTuple

from pyDE1.dispatcher.resource import Resource, RESOURCE_VERSION, \
    DE1ModeEnum, ConnectivityEnum

# TODO: Work through main and remote thread imports

logger = logging.getLogger(multiprocessing.current_process().name)

pname = f"into {multiprocessing.current_process().name} process"

if 'pyDE1.de1' in sys.modules:
    from pyDE1.de1 import DE1
else:
    print("Importing stub for DE1", pname)
    logger.info("Importing stub for DE1")
    from pyDE1.dispatcher.stubs import DE1

if 'pyDE1.scale' in sys.modules:
    from pyDE1.scale import Scale
else:
    print("Importing stub for Scale", pname)
    logger.info("Importing stub for Scale")
    from pyDE1.dispatcher.stubs import Scale

if 'pyDE1.flow_sequencer' in sys.modules:
    from pyDE1.flow_sequencer import FlowSequencer
else:
    print("Importing stub for FlowSequencer", pname)
    logger.info("Importing stub for FlowSequencer")
    from pyDE1.dispatcher.stubs import FlowSequencer

# cpn = multiprocessing.current_process().name
# for k in sys.modules.keys():
#     if (k.startswith('pyDE1')
#             or k.startswith('bleak')
#             or k.startswith('asyncio-mqtt')):
#         print(
#             f"{cpn}: EARLY: {k}"
#         )

from pyDE1.de1.c_api import PackedAttr, MMR0x80LowAddr, get_cuuid, \
    ShotSettings, SetTime, Versions, WaterLevels

from pyDE1.de1.exceptions import DE1APIValueError

MAPPING_VERSION = "2.1.0"


class IsAt (NamedTuple):
    target: Union[type(DE1),
                  type(Scale),
                  type(FlowSequencer),
                  type(PackedAttr),
                  type(MMR0x80LowAddr)]
    attr_path: Optional[str]  # '' for MMR0x80LowAddr, None for write_only,
    v_type: type  # expected value type for type-checking
    setter_path: Optional[str] = None   # If not a property and a different path
                                        # path of setter relative to target
    read_only: Optional[bool] = False
    internal_type: Optional[Union[type(DE1ModeEnum),
                                  type(ConnectivityEnum)]] = None
    # write_only: attr_path=None and setter_path=setter_attribute_name

    @property
    def requires_connected_de1(self) -> bool:
        return (
            self.target in (PackedAttr, MMR0x80LowAddr) or
            (self.target is DE1
             and self.setter_path not in (None, 'connectivity_setter'))
        )

    @property
    def requires_connected_scale(self) -> bool:
        return (
                self.target is Scale
                and self.setter_path not in (None, 'connectivity_setter')
        )


def mapping_requires(mapping: dict) -> dict:
    results = {
        'DE1': False,
        'Scale': False
    }
    return _mapping_requires_inner(mapping, results)


def _mapping_requires_inner(mapping: dict, results: dict) -> dict:

    for val in mapping.values():
        if isinstance(val, IsAt):
            if val.requires_connected_de1:
                results['DE1'] = True
            if val.requires_connected_scale:
                results['Scale'] = True

        if isinstance(val, dict):
            _mapping_requires_inner(val, results)

        # TODO: Maybe one day generify this
        if results['DE1'] and results['Scale']:
            break

    return results


def from_packed_attr(packed_attr: PackedAttr, attr_path: str, v_type: type,
                     setter_path: Optional[str] = None):
    # get_cuuid will raise if not over-the-wire
    cuuid = get_cuuid(packed_attr)
    can_read = cuuid.can_read or cuuid.can_notify
    can_write = cuuid.can_write
    if attr_path is not None and not can_read:
        raise DE1APIValueError(
            f"Un-readable PackedAttr {packed_attr} "
            f"with attr_path set to '{attr_path}'"
        )

    return IsAt(
        target= packed_attr,
        attr_path= attr_path,
        v_type= v_type,
        setter_path= setter_path,
        read_only= not can_read,
    )


def from_mmr(mmr: MMR0x80LowAddr, v_type: type):
    # This doesn't handle can't read and can't write
    return IsAt(
        target= mmr,
        attr_path= '',
        v_type= v_type,
        read_only= not mmr.can_read
    )


MODULES_FOR_VERSIONS = (
    'pyDE1',
    'bleak',
    'asyncio-mqtt',
    'paho-mqtt'
)


import importlib.metadata  # Used for module-version lookup only

def module_versions():
    retval = {}
    for module in MODULES_FOR_VERSIONS:
        retval[module] = importlib.metadata.version(module)
    return retval


MAPPING = dict()

MAPPING[Resource.VERSION] = {
    'resource_version': RESOURCE_VERSION,
    'mapping_version': MAPPING_VERSION,
    'platform': sys.platform,
    'python': sys.version,
    'python_info': {
        'major': sys.version_info.major,
        'minor': sys.version_info.minor,
        'micro': sys.version_info.micro,
        'releaselevel': sys.version_info.releaselevel,
        'serial': sys.version_info.serial,
    },
    'module_versions': module_versions()
}

# "Specials" -- content-only

MAPPING[Resource.DE1_PROFILE] = IsAt(target=DE1, attr_path=None,
                                     setter_path='upload_json_v2_profile',
                                     v_type=Union[bytes, bytearray])

# Work from leaves back, so can be "included" by reference

# For now, these are not writable
MAPPING[Resource.DE1_ID] = {
    'name': IsAt(target=DE1, attr_path='name', v_type=str,
                 read_only=True),
    'id': IsAt(target=DE1, attr_path='address', v_type=str,
               read_only=True),
}

# NB: A single-entry tuple needs to end with a comma

MAPPING[Resource.DE1_MODE] = {
    'mode': IsAt(target=DE1, attr_path=None, setter_path='mode_setter',
                 v_type=str, internal_type=DE1ModeEnum),
}
# TODO: de1.mode()

DE1_PROFILE = 'de1/profile'
# DE1_PROFILES = 'de1/profiles'
# DE1_PROFILE_UPLOAD = 'de1/profile/{id}/upload'

# DE1_FIRMWARE = 'de1/firmware'
# DE1_FIRMWARES = 'de1/firmwares'
# DE1_FIRMWARE_UPLOAD = 'de1/firmware/{id}/upload'

MAPPING[Resource.DE1_CONNECTIVITY] = {
    'mode': IsAt(target=DE1, attr_path='connectivity',
                 setter_path='connectivity_setter', v_type=str,
                 internal_type=ConnectivityEnum),
}

# DE1_CONTROL = 'de1/control' -- aggregate

# TODO: Work through how to get this to work

# TODO: Allow stop_at_xxxx for everything, None means don't apply

MAPPING[Resource.DE1_CONTROL_ESPRESSO] = {
    'stop_at_time': IsAt(target=FlowSequencer,
                         attr_path='espresso_control.stop_at_time',
                         v_type=Optional[float]),
    'stop_at_volume': IsAt(target=FlowSequencer,
                           attr_path='espresso_control.stop_at_volume',
                           v_type=Optional[float]),
    'stop_at_weight': IsAt(target=FlowSequencer,
                           attr_path='espresso_control.stop_at_weight',
                           v_type=Optional[float]),
    'disable_auto_tare': IsAt(target=FlowSequencer,
                              attr_path='espresso_control.disable_auto_tare',
                              v_type=bool),

    'profile_can_override_stop_limits':
        IsAt(target=FlowSequencer,
             attr_path='espresso_control.profile_can_override_stop_limits',
             v_type=bool),
    'profile_can_override_tank_temperature':
        IsAt(target=FlowSequencer,
             attr_path='espresso_control.profile_can_override_tank_temperature',
             v_type=bool),
    'first_drops_threshold':
        IsAt(target=FlowSequencer,
             attr_path='espresso_control.first_drops_threshold',
             v_type=Optional[float]),
    'last_drops_minimum_time':
        IsAt(target=FlowSequencer,
             attr_path='espresso_control.last_drops_minimum_time',
             v_type=float),
}

MAPPING[Resource.DE1_CONTROL_STEAM] = {
    'stop_at_time': IsAt(target=ShotSettings, attr_path='TargetSteamLength', v_type=int),
    'stop_at_volume': IsAt(target=FlowSequencer,
                           attr_path='steam_control.stop_at_volume',
                           v_type=Optional[float]),
    'stop_at_weight': IsAt(target=FlowSequencer,
                           attr_path='steam_control.stop_at_weight',
                           v_type=Optional[float]),
    'disable_auto_tare': IsAt(target=FlowSequencer,
                              attr_path='steam_control.disable_auto_tare',
                              v_type=bool),
}

MAPPING[Resource.DE1_CONTROL_HOT_WATER] = {
    'stop_at_time': IsAt(target=ShotSettings,
                         attr_path='TargetHotWaterLength',
                         v_type=int),
    'stop_at_volume': IsAt(target=ShotSettings,
                           attr_path='TargetHotWaterVol',
                           v_type=int),
    'stop_at_weight': IsAt(target=FlowSequencer,
                           attr_path='hot_water_control.stop_at_weight',
                           v_type=Optional[float]),
    'disable_auto_tare': IsAt(target=FlowSequencer,
                              attr_path='hot_water_control.disable_auto_tare',
                              v_type=bool),
    'temperature': IsAt(target=ShotSettings,
                        attr_path='TargetHotWaterTemp', v_type=int),
}

MAPPING[Resource.DE1_CONTROL_HOT_WATER_RINSE] = {
    'stop_at_time':
        IsAt(target=FlowSequencer,
             attr_path='hot_water_rinse_control.stop_at_time',
             v_type=Optional[float]),
    'stop_at_volume':
        IsAt(target=FlowSequencer,
             attr_path='hot_water_rinse_control.stop_at_volume',
             v_type=Optional[float]),
    'stop_at_weight':
        IsAt(target=FlowSequencer,
             attr_path='hot_water_rinse_control.stop_at_weight',
             v_type=Optional[float]),
    'disable_auto_tare':
        IsAt(target=FlowSequencer,
             attr_path='hot_water_rinse_control.disable_auto_tare', v_type=bool),
}

MAPPING[Resource.DE1_CONTROL_TANK_WATER_THRESHOLD] = {
    'temperature': IsAt(target=MMR0x80LowAddr.TANK_WATER_THRESHOLD, attr_path='', v_type=int),
}

# DE1_SETTING = 'de1/setting' -- aggregate

MAPPING[Resource.DE1_SETTING_AUTO_OFF_TIME] = {
    'time': IsAt(target=DE1, attr_path='auto_off_time',
                 v_type=Optional[float]),
}

MAPPING[Resource.DE1_SETTING_FAN_THRESHOLD] = {
    'temperature': IsAt(target=MMR0x80LowAddr.FAN_THRESHOLD, attr_path='', v_type=int),
}

MAPPING[Resource.DE1_SETTING_START_FILL_LEVEL] = {
    'start_fill_level': IsAt(target=WaterLevels, attr_path='StartFillLevel', v_type=int),
}

MAPPING[Resource.DE1_SETTING_BEFORE_FLOW] = {
    'heater_phase1_flow': IsAt(target=MMR0x80LowAddr.HEATER_PHASE1_FLOW, attr_path='', v_type=float),
    'heater_phase2_flow': IsAt(target=MMR0x80LowAddr.HEATER_PHASE2_FLOW, attr_path='', v_type=float),
    'heater_phase2_timeout': IsAt(target=MMR0x80LowAddr.HEATER_PHASE2_TIMEOUT, attr_path='', v_type=float),
    'heater_idle_temperature': IsAt(target=MMR0x80LowAddr.HEATER_IDLE_TEMPERATURE, attr_path='', v_type=float),
}

MAPPING[Resource.DE1_SETTING_STEAM] = {
    'temperature': IsAt(target=ShotSettings, attr_path='TargetSteamTemp', v_type=int),
    'flow': IsAt(target=MMR0x80LowAddr.STEAM_FLOW_RATE, attr_path='', v_type=float),
    'high_flow_time': IsAt(target=MMR0x80LowAddr.HIGH_STEAM_FLOW_TIME, attr_path='', v_type=float),
}

# TODO: What is ShotSettings.TargetGroupTemp and where does it really belong?
MAPPING[Resource.DE1_SETTING_TARGET_GROUP_TEMP] = {
    'temperature': IsAt(target=ShotSettings, attr_path='TargetGroupTemp', v_type=int),
}

# NB: Not exposed at this time
MAPPING[Resource.DE1_DEPRECATED] = {
    'old_espresso_vol': IsAt(target=ShotSettings,
                             attr_path='TargetEspressoVol', v_type=int),
    'steam_fast_start': IsAt(target=ShotSettings,
                       attr_path='steam_setting_fast_start', v_type=bool),
    'steam_high_power': IsAt(target=ShotSettings,
                       attr_path='steam_setting_high_power', v_type=bool),
}

MAPPING[Resource.DE1_SETTING_TIME] = {
    'timestamp': IsAt(target=SetTime, attr_path='Timestamp', v_type=int),
}

# None exposed yet
MAPPING[Resource.DE1_PARAMETER_SET] = {
}

MAPPING[Resource.DE1_READ_ONCE] = {
    'hw_config_hexstr': IsAt(target=MMR0x80LowAddr.HW_CONFIG, attr_path='', v_type=str),
    'model_hexstr': IsAt(target=MMR0x80LowAddr.MODEL, attr_path='', v_type=str),
    'cpu_board_model': IsAt(target=MMR0x80LowAddr.CPU_BOARD_MODEL, attr_path='', v_type=float),
    'firmware_model': IsAt(target=MMR0x80LowAddr.FIRMWARE_MODEL, attr_path='', v_type=str),
    'firmware_build_number': IsAt(target=MMR0x80LowAddr.FIRMWARE_BUILD_NUMBER, attr_path='', v_type=int),
    'ghc_info': IsAt(target=MMR0x80LowAddr.GHC_INFO, attr_path='', v_type=str),  # See MMRGHCInfoBitMask
    'serial_number_hexstr': IsAt(target=MMR0x80LowAddr.SERIAL_NUMBER, attr_path='', v_type=str),
    'heater_voltage': IsAt(target=MMR0x80LowAddr.HEATER_VOLTAGE, attr_path='', v_type=int),

    'version_ble': {
        'api': IsAt(target=Versions, attr_path='BLEVersion.APIVersion', v_type=int),
        'release': IsAt(target=Versions, attr_path='BLEVersion.Release', v_type=float),
        'commits': IsAt(target=Versions, attr_path='BLEVersion.Commits', v_type=int),
        'changes': IsAt(target=Versions, attr_path='BLEVersion.Changes', v_type=int),
        'blesha_hexstr': IsAt(target=Versions, attr_path='BLEVersion.BLESha', v_type=str),
    },

    'version_lv': {
        'api': IsAt(target=Versions, attr_path='LVVersion.APIVersion', v_type=int),
        'release': IsAt(target=Versions, attr_path='LVVersion.Release', v_type=float),
        'commits': IsAt(target=Versions, attr_path='LVVersion.Commits', v_type=int),
        'changes': IsAt(target=Versions, attr_path='LVVersion.Changes', v_type=int),
        'blesha_hexstr': IsAt(target=Versions, attr_path='LVVersion.BLESha', v_type=str),
    },
}

MAPPING[Resource.DE1_CALIBRATION_FLOW_MULTIPLIER] = {
    'multiplier': IsAt(target=MMR0x80LowAddr.FLOW_CALIBRATION,
                       attr_path='', v_type=float),
}

MAPPING[Resource.DE1_CALIBRATION_LINE_FREQUENCY] = {
    'hz': IsAt(target=DE1, attr_path='line_frequency', v_type=int)
}

MAPPING[Resource.SCALE_ID] = {
    'name': IsAt(target=Scale, attr_path='name', v_type=str,
                 read_only=True),
    'id': IsAt(target=Scale, attr_path='address', v_type=str,
               read_only=True),
    'type': IsAt(target=Scale, attr_path='type', v_type=str,
                 read_only=True),
}

MAPPING[Resource.SCALE_CONNECTIVITY] = {
    'mode': IsAt(target=Scale, attr_path='connectivity',
                 setter_path="connectivity_setter", v_type=str,
                 internal_type=ConnectivityEnum),
}

MAPPING[Resource.SCALE_TARE] = {
    'tare': IsAt(target=Scale, attr_path='', setter_path='tare_with_bool',
                 v_type=Optional[bool])     # Accommodate None as False
}

MAPPING[Resource.SCALE_DISPLAY] = {
    'display_on': IsAt(target=Scale, attr_path='', setter_path='display_bool',
                       v_type=Optional[bool])  # Accommodate None as False
}

MAPPING[Resource.DE1_CONTROL] = {
    'espresso': MAPPING[Resource.DE1_CONTROL_ESPRESSO],
    'steam': MAPPING[Resource.DE1_CONTROL_STEAM],
    'hot_water': MAPPING[Resource.DE1_CONTROL_HOT_WATER],
    'hot_water_rinse': MAPPING[Resource.DE1_CONTROL_HOT_WATER_RINSE],
    'tank_water_threshold': MAPPING[Resource.DE1_CONTROL_TANK_WATER_THRESHOLD],
}

MAPPING[Resource.DE1_SETTING] = {
    'auto_off_time': MAPPING[Resource.DE1_SETTING_AUTO_OFF_TIME],
    'fan_threshold': MAPPING[Resource.DE1_SETTING_FAN_THRESHOLD],
    'start_fill_level': MAPPING[Resource.DE1_SETTING_START_FILL_LEVEL],
    'before_flow': MAPPING[Resource.DE1_SETTING_BEFORE_FLOW],
    'target_group_temp': MAPPING[Resource.DE1_SETTING_TARGET_GROUP_TEMP],
    'steam': MAPPING[Resource.DE1_SETTING_STEAM],
    'time': MAPPING[Resource.DE1_SETTING_TIME],
}

MAPPING[Resource.DE1_CALIBRATION] = {
    'flow_multiplier': MAPPING[Resource.DE1_CALIBRATION_FLOW_MULTIPLIER],
    'line_frequency': MAPPING[Resource.DE1_CALIBRATION_LINE_FREQUENCY],
    # 'internal': Mapping[Resource.DE1_CALIBRATION_INTERNAL],
}

# TODO: How to handle GET if non-GET items? NaN? null?
MAPPING[Resource.DE1] = {
    'id': MAPPING[Resource.DE1_ID],
    'mode': MAPPING[Resource.DE1_MODE],
    # profile
    # profiles
    # firmware
    # firmwares
    'connectivity': MAPPING[Resource.DE1_CONNECTIVITY],
    'control': MAPPING[Resource.DE1_CONTROL],
    'setting': MAPPING[Resource.DE1_SETTING],
    'calibration': MAPPING[Resource.DE1_CALIBRATION],
    'parameter_set': MAPPING[Resource.DE1_PARAMETER_SET],
    'read_once': MAPPING[Resource.DE1_READ_ONCE],
}


MAPPING[Resource.SCALE] = {
    'id': MAPPING[Resource.SCALE_ID],
    'connectivity': MAPPING[Resource.SCALE_CONNECTIVITY],
    'tare': MAPPING[Resource.SCALE_TARE],
    'display': MAPPING[Resource.SCALE_DISPLAY],
}

MAPPING[Resource.FLOW_SEQUENCER_SETTING] = {}

MAPPING[Resource.FLOW_SEQUENCER_PARAMETER_SET] = {}

MAPPING[Resource.FLOW_SEQUENCER] = {
    'setting': MAPPING[Resource.FLOW_SEQUENCER_SETTING],
    'parameter_set': MAPPING[Resource.FLOW_SEQUENCER_PARAMETER_SET],
}



"""
Copyright © 2021 Jeff Kletsky. All Rights Reserved.

License for this software, part of the pyDE1 package, is granted under
GNU General Public License v3.0 only
SPDX-License-Identifier: GPL-3.0-only

"Null" outbound processor that just keeps the queue clear,
counting the various messages by type,
logging every update_period seconds (and resetting the count)
"""


# Only import the minimal here, as it potentially ends up in all processes.
import multiprocessing, multiprocessing.connection

# TODO: look into how loggers here relate to the root logger from "main"

# TODO: Look into or resolve processes' loggers writing over each other


def run_api_outbound(api_outbound_queue: multiprocessing.Queue):

    import logging
    import os
    import time

    from socket import gethostname

    logger = logging.getLogger(multiprocessing.current_process().name)

    import asyncio
    import json
    import sys

    # cpn = multiprocessing.current_process().name
    # for k in sys.modules.keys():
    #     if (k.startswith('pyDE1')
    #             or k.startswith('bleak')
    #             or k.startswith('asyncio-mqtt')):
    #         print(
    #             f"{cpn}: {k}"
    #         )

    import asyncio_mqtt
    import asyncio_mqtt.client

    # https://github.com/eclipse/paho.mqtt.c/issues/864
    # Add support for Unix-domain sockets #864 (open issue)

    # TODO: Move these into a settings object

    MQTT_TOPIC_ROOT = 'pyDE1'

    MQTT_CLIENT_ID = f"pyDE1@{gethostname()}[{os.getpid()}]"

    MQTT_BROKER_HOSTNAME = '::'
    MQTT_BROKER_PORT = 1883

    MQTT_TRANSPORT = 'tcp'
    MQTT_TLS_CONTEXT = None
    MQTT_KEEPALIVE = 60

    MQTT_USERNAME = None
    MQTT_PASSWORD = None

    MQTT_PROTOCOL_VERSION = asyncio_mqtt.client.ProtocolVersion.V5

    async def run(api_outbound_queue: multiprocessing.Queue):

        last_update = time.time()
        update_period = 10  # in seconds
        counts = {}

        async with asyncio_mqtt.client.Client(
            hostname=MQTT_BROKER_HOSTNAME,
            port=MQTT_BROKER_PORT,
            transport=MQTT_TRANSPORT,
            tls_context=MQTT_TLS_CONTEXT,
            keepalive=MQTT_KEEPALIVE,
            protocol=MQTT_PROTOCOL_VERSION,
            username=MQTT_USERNAME,
            password=MQTT_PASSWORD,
            will=None,
            client_id=MQTT_CLIENT_ID,
        ) as client:

            logger.info("Connecting to client")
            asyncio.run_coroutine_threadsafe(client.connect(),
                                             asyncio.get_running_loop())
            logger.info("Client connected")

            while True:
                item_json = api_outbound_queue.get()
                item_as_dict = json.loads(item_json)
                topic = f"{MQTT_TOPIC_ROOT}/{item_as_dict['class']}"
                await client.publish(
                    topic=topic,
                    payload=item_json,
                    qos=0,              # 0: At most once
                    retain=True,        # Always have "last known" available
                    properties=None,    # MQTT 5 only
                )
                now = time.time()
                try:
                    counts[item_as_dict['class']] += 1
                except KeyError:
                    counts[item_as_dict['class']] = 1
                if now - last_update > update_period:
                    logger.info(counts)
                    counts = {}
                    last_update = now

    loop = asyncio.get_event_loop()
    loop.set_debug(True)
    # loop.run_until_complete(run(api_outbound_queue=api_outbound_queue))
    loop.run_until_complete(run(api_outbound_queue=api_outbound_queue))



"""
Copyright © 2021 Jeff Kletsky. All Rights Reserved.

License for this software, part of the pyDE1 package, is granted under
GNU General Public License v3.0 only
SPDX-License-Identifier: GPL-3.0-only


FlowSequencer

Responsible for coordinating the actions around any of the flow modes
"""
import enum
import time

import asyncio
import logging

from typing import Optional, Callable, Coroutine, List

from pyDE1.de1 import DE1
from pyDE1.de1.events import ShotSampleUpdate, StateUpdate, \
    ShotSampleWithVolumesUpdate
from pyDE1.de1.c_api import API_MachineStates, API_Substates, CUUID
from pyDE1.de1.exceptions import DE1APIError, DE1APITypeError, DE1APIValueError, \
    DE1APIAttributeError

from pyDE1.i_target_setter import I_TargetSetter

from pyDE1.scale.processor import ScaleProcessor
from pyDE1.scale.events import WeightAndFlowUpdate

from pyDE1.event_manager import EventWithNotify, EventNotificationName, \
    EventPayload, send_to_outbound_queue
from pyDE1.singleton import Singleton

from pyDE1.utils import cancel_tasks_by_name

from pyDE1.flow_sequencer.mode_control import ByModeControl, \
    StopAtTime, StopAtWeight, StopAtVolume,  \
    EspressoControl, SteamControl, HotWaterControl, HotWaterRinseControl

logger = logging.getLogger('FlowSequencer')


# TODO: These didn't neatly become inner classes

class StopAtNotificationAction (enum.Enum):
    ENABLED = 'enabled'
    TRIGGERED = 'triggered'
    DISABLED = 'disabled'


class StopAtType (enum.Enum):
    TIME = 'time'
    VOLUME = 'volume'
    WEIGHT = 'weight'


class StopAtNotification (EventPayload):
    """
    Enable, disable, trigger of the various stop-at conditions
    current_value is generally only set for StopAtNotificationAction.TRIGGERED

    ENABLED notifications are given even if the target is None as the target
    can be changed during the shot, at least when managed by the FlowSequencer
    (On-the-fly profile changes are not supported at this time.
     On-the-fly changes to steam duration have not been tested at this time.)
    """
    def __init__(self, stop_at: StopAtType,
                 action: StopAtNotificationAction,
                 target_value: Optional[float] = None,
                 current_value: Optional[float] = None,
                 active_state: API_MachineStates = API_MachineStates.NoRequest):
        now = time.time()
        super(StopAtNotification, self).__init__(
            arrival_time=now,
            create_time=now
        )
        self._version = "1.0.0"
        self.stop_at = stop_at
        self.action = action
        self.target_value = target_value
        self.current_value = current_value
        self.active_state = active_state


class AutoTareNotificationAction (enum.Enum):
    ENABLED = 'enabled'
    DISABLED = 'disabled'


class AutoTareNotification (EventPayload):

    def __init__(self, action: AutoTareNotificationAction):
        now = time.time()
        super(AutoTareNotification, self).__init__(
            arrival_time=now,
            create_time=now
        )
        self._version = "1.0.0"
        self.action = action


class FlowSequencer (Singleton, I_TargetSetter):

    # NB: This is intentionally done in _singleton_init() and not __init__()
    #     See Singleton and Guido's notes there
    #
    # def __init__(self):
    #     pass

    def _singleton_init(self):
        self._de1: Optional[DE1] = None
        self._scale_processor: Optional[ScaleProcessor] = None

        self._event_de1_present = asyncio.Event()
        self._event_scale_processor_present = asyncio.Event()

        self._active_state: Optional[API_MachineStates] = None
        self._sequence_start_time: float = 0
        # For later cancellation. Should this be static or dynamic?
        self._sequence_task_name: str = "SequencerSubtask"

        self._gate_sequence_start = EventWithNotify(self,
            EventNotificationName.GATE_SEQUENCE_START)
        # These are not necessarily in order
        self._gate_flow_begin = EventWithNotify(self,
            EventNotificationName.GATE_FLOW_BEGIN)
        self._gate_expect_drops = EventWithNotify(self,
            EventNotificationName.GATE_EXPECT_DROPS)
        # self._gate_first_drops = asyncio.Event()
        self._gate_exit_preinfuse = EventWithNotify(self,
            EventNotificationName.GATE_EXIT_PREINFUSE)
        self._gate_flow_end = EventWithNotify(self,
            EventNotificationName.GATE_FLOW_END)
        self._gate_flow_state_exit = EventWithNotify(self,
            EventNotificationName.GATE_FLOW_STATE_EXIT)
        self._gate_last_drops = EventWithNotify(self,
            EventNotificationName.GATE_LAST_DROPS)
        # Though always "done" here (and should always be triggered)
        self._gate_sequence_complete = EventWithNotify(self,
            EventNotificationName.GATE_SEQUENCE_COMPLETE)

        self._all_gates = [
            self._gate_sequence_start,
            self._gate_flow_begin,
            self._gate_expect_drops,
            self._gate_exit_preinfuse,
            self._gate_flow_end,
            self._gate_flow_state_exit,
            self._gate_last_drops,
            self._gate_sequence_complete,
        ]

        self.espresso_control = EspressoControl()
        self.steam_control = SteamControl()
        self.hot_water_control = HotWaterControl()
        self.hot_water_rinse_control = HotWaterRinseControl()

        # TODO: Refactor these out in favor of None in the ByModeControl
        self.autotare_states: List[API_MachineStates] = [
            API_MachineStates.Espresso,
            API_MachineStates.HotWater,
        ]

        self.stop_at_weight_states: List[API_MachineStates] = [
            API_MachineStates.Espresso,
            API_MachineStates.HotWater,
        ]

        self.stop_at_volume_states: List[API_MachineStates] = [
            API_MachineStates.Espresso,
            API_MachineStates.HotWater,
        ]

        self.last_drops_states = [
            API_MachineStates.Espresso,
            API_MachineStates.HotWater,     # TODO: Needed? Adjust?
            API_MachineStates.Steam,        # TODO: Needed? Adjust?
        ]

        self.stop_at_time_states = [
            API_MachineStates.HotWaterRinse,
            # API_MachineStates.Steam,  # Handled through ShotSettings
        ]

        # Recorder gets managed by the recorder itself

        # Internal flag, use None or value for
        # de1.stop_at_weight and/or de1.stop_at_volume
        self._stop_at_weight_active = False
        self._stop_at_volume_active = False
        self._stop_at_time_active = False

    @property
    def de1(self):
        return self._de1

    async def set_de1(self, value: DE1):
        # Needs to unsubscribe the old
        self._de1 = value
        self._de1._flow_sequencer = self
        self._event_de1_present.set()
        await asyncio.gather(
            self.de1.event_state_update.subscribe(
                self._create_state_update_callback()),

            self.de1.event_shot_sample.subscribe(
                self._create_shot_sample_update_callback()),

            self.de1.event_shot_sample_with_volumes_update.subscribe(
                self._create_stop_at_volume_subscriber()),

            # This will fail if the scale_processor hasn't been connected yet
            # self.scale_processor.event_weight_and_flow_update.subscribe(
            #     _create_stop_at_weight_subscriber(self.de1)),
        )
        asyncio.create_task(self._set_stop_at_weight_subscriber_when_possible())
        return self

    #
    # TODO: THIS NEEDS A LOT OF HELP FOR CHANGING EITHER
    #
    async def _set_stop_at_weight_subscriber_when_possible(self):
        await asyncio.gather(
            self._event_scale_processor_present.wait(),
            self._event_de1_present.wait(),
        )
        await self.scale_processor.event_weight_and_flow_update.subscribe(
            self._create_stop_at_weight_subscriber()
        )

    @property
    def scale_processor(self):
        return self._scale_processor

    async def set_scale_processor(self, value: ScaleProcessor):
        # Needs to unsubscribe the old
        self._scale_processor = value
        self._event_scale_processor_present.set()
        await self._scale_processor.event_weight_and_flow_update.subscribe(
                self._create_weight_and_flow_update_callback())
        return self

    @property
    def active_state(self):
        return self._active_state

    def active_control_for_state(self, state: API_MachineStates) \
            -> ByModeControl:
        retval = None
        if state is API_MachineStates.Espresso:
            retval = self.espresso_control
        elif state is API_MachineStates.Steam:
            retval = self.steam_control
        elif state is API_MachineStates.HotWater:
            retval = self.hot_water_control
        elif state is API_MachineStates.HotWaterRinse:
            retval = self.hot_water_rinse_control
        return retval

    @property
    def active_control(self) -> ByModeControl:
        return self.active_control_for_state(self.active_state)

    @property
    def sequence_start_time(self):
        return self._sequence_start_time

    def _close_all_gates(self):
        for gate in self._all_gates:
            gate.clear()

    # Weight and Flow Update handles auto-tare
    # SAW and SAV have their own handlers, supervisor enables and disables
    # Subscribe to the ScaleProcessor, let it handle scale-instance changes

    def _create_weight_and_flow_update_callback(self) -> Callable:
        flow_sequencer = self

        async def flow_sequencer_wafu_cb(wafu: WeightAndFlowUpdate):
            nonlocal flow_sequencer
            pass  # at least now, all the SAW is in ScaleProcessor

        return flow_sequencer_wafu_cb

    def _create_state_update_callback(self) -> Callable:
        flow_sequencer = self

        async def flow_sequencer_su_cb(su: StateUpdate):
            nonlocal flow_sequencer

            if su.state.is_flow_state:

                if not su.previous_state.is_flow_state:
                    flow_sequencer._start_sequence(su.state)
                    logger.info("Start sequence")

                if su.substate.flow_phase == 'during' \
                        and su.previous_substate.flow_phase != 'during':
                    flow_sequencer._gate_flow_begin.set()
                    logger.info("Gate: Flow begin")

                if su.previous_substate == API_Substates.PreInfuse \
                        and su.substate != API_Substates.PreInfuse:
                    flow_sequencer._gate_exit_preinfuse.set()
                    logger.info("Gate: Exit preinfuse")

                if su.previous_substate.flow_phase == 'during' \
                        and su.substate.flow_phase != 'during':
                    flow_sequencer._gate_flow_end.set()
                    asyncio.create_task(self._wait_for_last_drops(),
                                        name=self._sequence_task_name)
                    logger.info("Gate: Flow end")

            elif su.previous_state.is_flow_state:  # current is not a flow state
                if su.previous_substate == API_Substates.PreInfuse:
                    flow_sequencer._gate_exit_preinfuse.set()
                    logger.info("Gate: Exit preinfuse")
                if su.previous_substate.flow_phase == 'during':
                    flow_sequencer._gate_flow_end.set()
                    logger.info("Gate: Flow end")
                flow_sequencer._gate_flow_state_exit.set()
                logger.info("Gate: Flow state exit")

        return flow_sequencer_su_cb

    def _create_shot_sample_update_callback(self) -> Callable:
        flow_sequencer = self

        async def flow_sequencer_ssu_cb(ssu: ShotSampleUpdate):
            nonlocal flow_sequencer

            # Yet another place having to use current [sub]state off DE1
            de1 = flow_sequencer.de1
            try:
                threshold = flow_sequencer.active_control.first_drops_threshold
            except AttributeError:
                threshold = None
            if de1.current_state.is_flow_state \
                    and threshold is not None \
                    and not flow_sequencer._gate_expect_drops.is_set() \
                    and de1.current_substate.flow_phase == 'during' \
                    and ssu.group_pressure \
                        > flow_sequencer.active_control.first_drops_threshold:
                flow_sequencer._gate_expect_drops.set()
                logger.info("Gate: Expect drops")

        return flow_sequencer_ssu_cb

    def _start_sequence(self, state: API_MachineStates):
        """
        Kick off parallel tasks to manage functions during the shot
        Tasks should wait for self._gate_sequence_start
        """
        self._sequence_start_time = time.time()
        # TODO: Is there a more robust way to transition from an "aborted"
        #       shot to another one?
        if not self._gate_sequence_complete.is_set() \
                and self._sequence_start_time != 0:
            self._abort_sequence()

        self._active_state = state
        self._close_all_gates()

        # create a bunch of tasks
        # Keep the start gate closed, as some may expect others to exist
        # or need to
        # For now, don't name uniquely and see what the logger returns

        asyncio.create_task(self._sequence_end_sequence(),
                            name=self._sequence_task_name)

        asyncio.create_task(self._sequence_volume_tracking(),
                            name=self._sequence_task_name)

        asyncio.create_task(self._sequence_stop_at_volume(),
                            name=self._sequence_task_name)

        asyncio.create_task(self._sequence_scale_tare(),
                            name=self._sequence_task_name)

        asyncio.create_task(self._sequence_stop_at_weight(),
                            name=self._sequence_task_name)

        asyncio.create_task(self._sequence_stop_at_time(),
                            name=self._sequence_task_name)

        asyncio.create_task(self._sequence_recorder(),
                            name=self._sequence_task_name)

        self._gate_sequence_start.set()
        logger.info("Gate: Sequence start")

    async def _wait_for_last_drops(self):
        ldmt = self.active_control.last_drops_minimum_time
        if ldmt:
            await asyncio.sleep(ldmt)
        self._gate_last_drops.set()
        logger.info("Gate: Last drops")

    async def _sequence_end_sequence(self):
        """
        Maybe a funky way to do it, but this captures "done-ness"
        in one place.
        """
        await self._gate_flow_state_exit.wait()
        ldmt = self.active_control.last_drops_minimum_time
        if ldmt:
            try:
                await asyncio.wait_for(self._gate_last_drops.wait(),
                                  timeout=ldmt)
            except asyncio.exceptions.TimeoutError:
                pass
        await self._end_sequence()

    async def _end_sequence(self):
        """
        Not all subtasks will have gone through all their gates
        Set the completion Event and give a second for tasks to complete
        Clean up any stragglers
        """
        self._gate_sequence_complete.set()
        logger.info("Gate: Sequence complete")
        await asyncio.sleep(0.100)
        self._cleanup_stragglers()

    def _abort_sequence(self):
        """
        Not all subtasks will have gone through all their gates
        or sequence_complete may not have been set.

        This is a "no-wait" version

        Set the completion Event
        Clean up any stragglers
        """
        self._gate_sequence_complete.set()
        logger.info("Gate: Sequence complete: _abort_sequence()")
        self._cleanup_stragglers()

    def _cleanup_stragglers(self):
        """
        This will send CancelledError to the coroutine

        That should be caught and handled for cleanup
        or completion of tasks, such as writing files

        cancel_tasks_by_name() won't cancel its own task
        """
        logger.info("Canceling stragglers")
        cancel_tasks_by_name(self._sequence_task_name, starts_with=False)


###
### Sequences
###

    # TODO: These should be sensitive to the kind of flow
    #       just get them in for espresso for now

    async def _sequence_volume_tracking(self):
        de1 = self.de1
        try:
            await self._gate_sequence_start.wait()
            de1._reset_volume_dispensed()
            logger.debug("Volume: reset")

            await self._gate_flow_begin.wait()
            de1._tracking_volume_dispensed = True
            logger.debug("Volume: tracking start")

            await self._gate_flow_end.wait()
            de1._tracking_volume_dispensed = False
            logger.debug("Volume: tracking stop")

        except asyncio.CancelledError:
            de1._tracking_volume_dispensed = False
            logger.info("Volume: tracking stop - on cancel")
            raise

    async def _sequence_scale_tare(self):
        scale = self._scale_processor.scale
        if self.active_state not in self.autotare_states:
            scale.hold_at_tare = False
            logger.debug(f"Scale: release - {self.active_state.name}")
            await send_to_outbound_queue(AutoTareNotification(
                AutoTareNotificationAction.DISABLED))
            return
        try:
            await self._gate_sequence_start.wait()
            scale.hold_at_tare = True
            logger.debug("Scale: hold at tare")
            await send_to_outbound_queue(AutoTareNotification(
                AutoTareNotificationAction.ENABLED))

            await self._gate_expect_drops.wait()
            scale.hold_at_tare = False
            logger.debug("Scale: release")
            await send_to_outbound_queue(AutoTareNotification(
                AutoTareNotificationAction.DISABLED))

        except asyncio.CancelledError:
            scale = self._scale_processor.scale
            scale.hold_at_tare = False
            logger.info("Scale: release - on cancel")
            await send_to_outbound_queue(AutoTareNotification(
                AutoTareNotificationAction.DISABLED))
            raise

    async def _sequence_stop_at_volume(self):
        if self.active_state not in self.stop_at_volume_states:
            self._stop_at_volume_active = False
            logger.debug(f"StopAtVolume: disable - {self.active_state.name}")
            await self.stop_at_notify(
                stop_at=StopAtType.VOLUME,
                action=StopAtNotificationAction.DISABLED)
            return
        try:
            await self._gate_sequence_start.wait()
            self._stop_at_volume_active = False
            logger.debug("StopAtVolume: disable")
            await self.stop_at_notify(
                stop_at=StopAtType.VOLUME,
                action=StopAtNotificationAction.DISABLED)

            await self._gate_exit_preinfuse.wait()
            self._stop_at_volume_active = True
            logger.debug("StopAtVolume: enable")
            await self.stop_at_notify(
                stop_at=StopAtType.VOLUME,
                action=StopAtNotificationAction.ENABLED)

            await self._gate_flow_end.wait()
            self._stop_at_volume_active = False
            logger.debug("StopAtVolume: disable")
            await self.stop_at_notify(
                stop_at=StopAtType.VOLUME,
                action=StopAtNotificationAction.DISABLED)

        except asyncio.CancelledError:
            self._stop_at_volume_active = False
            logger.info("StopAtVolume: disable - on cancel")
            await self.stop_at_notify(
                stop_at=StopAtType.VOLUME,
                action=StopAtNotificationAction.DISABLED)
            raise

    async def _sequence_stop_at_weight(self):
        if self.active_state not in self.stop_at_weight_states:
            self._stop_at_weight_active = False
            logger.info(f"StopAtWeight: disable - {self.active_state.name}")
            await self.stop_at_notify(
                stop_at=StopAtType.WEIGHT,
                action=StopAtNotificationAction.DISABLED)
            return
        try:
            await self._gate_sequence_start.wait()
            self._stop_at_weight_active = False
            logger.debug("StopAtWeight: disable")
            await self.stop_at_notify(
                stop_at=StopAtType.WEIGHT,
                action=StopAtNotificationAction.DISABLED)

            await self._gate_expect_drops.wait()
            self._stop_at_weight_active = True
            logger.debug("StopAtWeight: enable")
            await self.stop_at_notify(
                stop_at=StopAtType.WEIGHT,
                action=StopAtNotificationAction.ENABLED)

            await self._gate_flow_end.wait()
            self._stop_at_weight_active = False
            logger.debug("StopAtWeight: disable")
            await self.stop_at_notify(
                stop_at=StopAtType.WEIGHT,
                action=StopAtNotificationAction.DISABLED)

        except asyncio.CancelledError:
            self._stop_at_weight_active = False
            logger.info("StopAtWeight: disable - on cancel")
            await self.stop_at_notify(
                stop_at=StopAtType.WEIGHT,
                action=StopAtNotificationAction.DISABLED)
            raise

    async def _sequence_stop_at_time(self):
        if self.active_state not in self.stop_at_time_states:
            self._stop_at_time_active = False
            logger.info(f"StopAtTime: disable - {self.active_state.name}")
            await self.stop_at_notify(
                stop_at=StopAtType.TIME,
                action=StopAtNotificationAction.DISABLED)
            return
        try:
            await self._gate_sequence_start.wait()
            self._stop_at_time_active = False
            logger.debug("StopAtTime: disable")
            await self.stop_at_notify(
                stop_at=StopAtType.TIME,
                action=StopAtNotificationAction.DISABLED)

            # NB: Changing the time after starting won't alter the duration
            #     at least as presently implemented

            wait = self.active_control.stop_at_time

            if wait is None or wait <= 0:
                return

            await self._gate_flow_begin.wait()
            self._stop_at_time_active = True
            t0 = time.time()
            logger.debug(f"StopAtTime: enable ({wait} seconds)")
            await self.stop_at_notify(
                stop_at=StopAtType.TIME,
                action=StopAtNotificationAction.ENABLED)

            await asyncio.sleep(wait)
            logger.debug(f"StopAtTime: triggered, requesting stop_flow")
            await self.de1.stop_flow()
            logger.debug(f"StopAtTime: triggered ({wait} seconds)")
            await self.stop_at_notify(
                stop_at=StopAtType.TIME,
                action=StopAtNotificationAction.TRIGGERED,
                current=(time.time() - t0))

            await self._gate_flow_end.wait()
            self._stop_at_time_active = False
            logger.debug("StopAtTime: disable")
            await self.stop_at_notify(
                stop_at=StopAtType.TIME,
                action=StopAtNotificationAction.DISABLED)

        except asyncio.CancelledError:
            self._stop_at_time_active = False
            logger.info("StopAtTime: disable - on cancel")
            await self.stop_at_notify(
                stop_at=StopAtType.TIME,
                action=StopAtNotificationAction.DISABLED)
            raise

    async def _sequence_recorder(self):
        de1 = self._de1
        try:
            # Always enable recording, let the recorder decide
            await self._gate_sequence_start.wait()
            de1._recorder_active = True
            logger.debug("Recorder: enable")

            await self._gate_sequence_complete.wait()
            de1._recorder_active = False
            logger.debug("Recorder: disable")

        except asyncio.CancelledError:
            de1._recorder_active = False
            logger.info("Recorder: disable - on cancel")
            raise

    def _create_stop_at_volume_subscriber(self) -> Coroutine:
        """
        Should be subscribed to ShotSampleWithVolumesUpdate on DE1
        """
        flow_sequencer = self
        sav_logger = logging.getLogger('StopAtVolume')

        async def stop_at_volume(sswvu: ShotSampleWithVolumesUpdate):
            nonlocal flow_sequencer, sav_logger

            if (flow_sequencer._stop_at_volume_active
                    and (target := flow_sequencer.active_control.stop_at_volume)
                    is not None):
                if sswvu.volume_pour >= target:
                    await flow_sequencer.de1.stop_flow()
                    sav_logger.info(
                        "Triggered at {:.1f} mL for {:.1f} mL target".format(
                            sswvu.volume_pour, target))
                    await flow_sequencer.stop_at_notify(
                        stop_at=StopAtType.VOLUME,
                        action=StopAtNotificationAction.TRIGGERED,
                        current=sswvu.volume_pour)

        return stop_at_volume

    def _create_stop_at_weight_subscriber(self) -> Coroutine:
        """
        Should be subscribed to WeightAndFlowUpdate on ScaleProcessor
        """
        flow_sequencer = self
        saw_logger = logging.getLogger('StopAtWeight')

        async def stop_at_weight(wafu: WeightAndFlowUpdate):
            nonlocal flow_sequencer, saw_logger

            if (flow_sequencer._stop_at_weight_active
                    and (target := flow_sequencer.active_control.stop_at_weight)
                    is not None):
                # TODO: where does this belong?
                adjust = -0.3  # seconds, larger means more in cup
                # TODO: where does this belong?
                stop_lag = flow_sequencer.de1.stop_lead_time
                # TODO: Should this be switchable?
                flow = wafu.average_flow
                dw = target - wafu.current_weight
                if flow > 0:
                    dt = dw / flow
                    target_time = wafu.scale_time + dt + adjust - stop_lag
                    if time.time() >= target_time:
                        await flow_sequencer.de1.stop_flow()
                        saw_logger.info(
                            "Triggered at {:.1f} g for {:.1f} g target".format(
                                wafu.current_weight, target))
                        await flow_sequencer.stop_at_notify(
                            stop_at=StopAtType.WEIGHT,
                            action=StopAtNotificationAction.TRIGGERED,
                            current=wafu.current_weight)

        return stop_at_weight

    async def stop_at_notify(self, stop_at: StopAtType,
                             action: StopAtNotificationAction,
                             current: Optional[float]=None):

        if stop_at == StopAtType.TIME:
            target = self.active_control.stop_at_time
        elif stop_at == StopAtType.VOLUME:
            target = self.active_control.stop_at_volume
        elif stop_at == StopAtType.WEIGHT:
            target = self.active_control.stop_at_weight
        else:
            target = None
        await send_to_outbound_queue(StopAtNotification(
            stop_at=stop_at,
            action=action,
            target_value=target,
            current_value=current,
            active_state=self.active_state
        ))

    # Implement I_TargetSetter

    # TODO: Handle profile lock. try/except AttributeError?

    def stop_at_time_set(self, state: API_MachineStates, duration: float):
        bmc = self.active_control_for_state(state)
        if bmc is None:
            raise DE1APIValueError(
                f"No ByModeController for {state}")
        if not isinstance(bmc, StopAtTime):
            raise DE1APIAttributeError(
                f"Not a StopAtTime ByModeControl {bmc}"
            )
        try:
            if not bmc.profile_can_override_stop_limits:
                return
        except AttributeError:
            pass
        bmc.stop_at_time = duration

    def stop_at_volume_set(self, state: API_MachineStates, volume: float):
        bmc = self.active_control_for_state(state)
        if bmc is None:
            raise DE1APIValueError(
                f"No ByModeController for {state}")
        if not isinstance(bmc, StopAtVolume):
            raise DE1APIAttributeError(
                f"Not a StopAtTime ByModeControl {bmc}"
            )
        try:
            if not bmc.profile_can_override_stop_limits:
                return
        except AttributeError:
            pass
        bmc.stop_at_volume = volume

    def stop_at_weight_set(self, state: API_MachineStates, weight: float):
        bmc = self.active_control_for_state(state)
        if bmc is None:
            raise DE1APIValueError(
                f"No ByModeController for {state}")
        if not isinstance(bmc, StopAtWeight):
            raise DE1APIAttributeError(
                f"Not a StopAtTime ByModeControl {bmc}"
            )
        try:
            if not bmc.profile_can_override_stop_limits:
                return
        except AttributeError:
            pass
        bmc.stop_at_weight = weight

    def profile_can_override_stop_limits(self, state: API_MachineStates):
        bmc = self.active_control_for_state(state)
        if bmc is None:
            raise DE1APIValueError(
                f"No ByModeController for {state}")
        try:
            return bmc.profile_can_override_stop_limits
        except AttributeError:
            return True

    def profile_can_override_tank_temperature(self, state: API_MachineStates):
        bmc = self.active_control_for_state(state)
        if bmc is None:
            raise DE1APIValueError(
                f"No ByModeController for {state}")
        try:
            return bmc.profile_can_override_tank_temperature
        except AttributeError:
            return True


"""
Copyright © 2021 Jeff Kletsky. All Rights Reserved.

License for this software, part of the pyDE1 package, is granted under
GNU General Public License v3.0 only
SPDX-License-Identifier: GPL-3.0-only
"""
import asyncio
import logging
import multiprocessing, multiprocessing.connection
import time

from pyDE1.de1 import DE1
from pyDE1.de1.profile import ProfileByFrames
from pyDE1.de1.exceptions import DE1NotConnectedError
from pyDE1.scale import Scale, AtomaxSkaleII
from pyDE1.flow_sequencer import FlowSequencer

from pyDE1.dispatcher.resource import Resource
from pyDE1.dispatcher.payloads import APIRequest, APIResponse, HTTPMethod
from pyDE1.dispatcher.implementation import get_resource_to_dict, \
    patch_resource_from_dict
from pyDE1.scale.processor import ScaleProcessor

READ_BACK_ON_PATCH = False

QUEUE_TOO_DEEP = 1  # If deeper than this, something is probably wrong, log

logger = logging.getLogger('Dispatcher')


# Interface from multiprocessing (sync) to asyncio.Queue() for async processing

def register_read_pipe_to_queue(
        pipe_to_read: multiprocessing.connection.Connection,
        queue_to_put: asyncio.Queue):
    asyncio.get_event_loop().add_reader(
        pipe_to_read.fileno(),
        _read_pipe_to_queue, pipe_to_read, queue_to_put)


def _read_pipe_to_queue(pipe_to_read: multiprocessing.connection.Connection,
                        queue_to_put: asyncio.Queue):
    data = pipe_to_read.recv()
    queue_to_put.put_nowait(data)
    if (qd := queue_to_put.qsize()) > QUEUE_TOO_DEEP:
        logger.error(
            f"Request queue exceeded QUEUE_TOO_DEEP, {qd} > {QUEUE_TOO_DEEP}")


def start_response_queue_processor(
        response_queue: asyncio.Queue,
        response_pipe: multiprocessing.connection.Connection):
    asyncio.create_task(_response_queue_processor(
        response_queue=response_queue,
        response_pipe=response_pipe),
    name='ResponseQueueProcessor')


async def _response_queue_processor(response_queue: asyncio.Queue,
        response_pipe: multiprocessing.connection.Connection):

    while True:
        response = await response_queue.get()
        response_pipe.send(response)


# Process received APIRequests from the inbound queue,
# write to the outbound queue.
# TODO: Move the outbound to an asyncio.Queue() later

def start_request_queue_processor(request_queue: asyncio.Queue,
                                  response_queue: asyncio.Queue):
    asyncio.create_task(_request_queue_processor(request_queue=request_queue,
                                                 response_queue=response_queue),
                        name='InboundQueueProcessor')


async def _request_queue_processor(request_queue: asyncio.Queue,
                                   response_queue: asyncio.Queue):

    flow_sequencer = FlowSequencer()
    de1 = DE1()
    scale_processor = ScaleProcessor()
    # scale = ScaleProcessor.scale  # No .scale when this starts

    while True:
        got: APIRequest = await request_queue.get()
        print(f"{type(got)}: {got.method} {got.resource} {got.connectivity_required}")
        resource_dict = {}
        exception = None
        if got.method is HTTPMethod.GET:
            try:
                # TODO: Should be "ready" and not just "connected"
                if not de1.is_connected \
                        and got.connectivity_required['DE1']:
                    raise DE1NotConnectedError("DE1 not connected")
                if not scale_processor.scale.is_connected \
                    and got.connectivity_required['Scale']:
                    raise DE1NotConnectedError("Scale not connected")
                resource_dict = await get_resource_to_dict(got.resource)
            except Exception as e:
                exception = e
                logger.error(
                    f"Exception in processing {got.method} {got.resource}"
                    f" {repr(exception)}")
            response = APIResponse(
                original_timestamp=got.timestamp,
                timestamp=time.time(),
                payload = resource_dict,
                exception=exception,
            )

        elif got.method is HTTPMethod.PATCH:
            resource_dict = None
            try:
                # TODO: Should be "ready" and not just "connected"
                if not de1.is_connected \
                        and got.connectivity_required['DE1']:
                    raise DE1NotConnectedError("DE1 not connected")
                if not scale_processor.scale.is_connected \
                    and got.connectivity_required['Scale']:
                    raise DE1NotConnectedError("Scale not connected")
                await patch_resource_from_dict(got.resource,
                                               got.payload)
            except Exception as e:
                exception = e
                logger.error(
                    f"Exception in processing {got.method} {got.resource}"
                    f" {repr(exception)}")
            if READ_BACK_ON_PATCH and \
                    exception is None and got.resource.can_get:
                resource_dict = await get_resource_to_dict(got.resource)
            response = APIResponse(
                original_timestamp=got.timestamp,
                timestamp=time.time(),
                payload = resource_dict,
                exception=exception,
            )

        elif got.method is HTTPMethod.PUT \
                and got.resource is Resource.DE1_PROFILE:
            try:
                # TODO: Should be "ready" and not just "connected"
                if not de1.is_connected \
                        and got.connectivity_required['DE1']:
                    raise DE1NotConnectedError("DE1 not connected")
                if not scale_processor.scale.is_connected \
                    and got.connectivity_required['Scale']:
                    raise DE1NotConnectedError("Scale not connected")

                # TODO: Implement PUT properly (check for completeness)
                await patch_resource_from_dict(got.resource,
                                               got.payload)
            except Exception as e:
                exception = e
                logger.error(
                    f"Exception in processing {got.method} {got.resource}"
                    f" {repr(exception)}")

            response = APIResponse(
                original_timestamp=got.timestamp,
                timestamp=time.time(),
                payload = None,
                exception=exception,
            )


        else:
            response = APIResponse(
                original_timestamp=got.timestamp,
                timestamp=time.time(),
                payload={},
                exception=NotImplementedError(
                    f"{got.method} is not supported"
                ),
            )

        response_queue.put_nowait(response)
        if (qd := response_queue.qsize()) > QUEUE_TOO_DEEP:
            logger.error(
                "Response queue exceeded QUEUE_TOO_DEEP, "
                f"{qd} > {QUEUE_TOO_DEEP}")



"""
Copyright © 2021 Jeff Kletsky. All Rights Reserved.

License for this software, part of the pyDE1 package, is granted under
GNU General Public License v3.0 only
SPDX-License-Identifier: GPL-3.0-only

Details of the payloads expected/supplied by each of the references

These should be pickle-able so can be passed between processes
"""

import enum
import json

from typing import Optional, Union

from pyDE1.dispatcher.resource import Resource

from pyDE1.de1.exceptions import DE1ValueError
from pyDE1.utils import fix_enums


class HTTPMethod (enum.Enum):

    GET = 'GET'
    PUT = 'PUT'
    PATCH = 'PATCH'

    HEAD = 'HEAD'
    POST = 'POST'
    DELETE = 'DELETE'
    OPTIONS = 'OPTIONS'
    TRACE = 'TRACE'
    CONNECT = 'CONNECT'

    def is_supported(self):
        return self in (
            self.GET,
            self.PUT,
            self.PATCH,
        )


class APIRequest:

    def __init__(self, timestamp: float,
                 method: HTTPMethod,
                 resource: Resource,
                 connectivity_required: dict,
                 payload,
                 ):
        self._timestamp = timestamp
        self._method = method
        self._resource = resource
        self._connectivity_required = connectivity_required
        self._payload = payload

    @property
    def timestamp(self):
        return self._timestamp

    @property
    def method(self):
        return self._method

    @property
    def resource(self):
        return self._resource

    @property
    def connectivity_required(self):
        return self._connectivity_required

    @property
    def payload(self):
        return self._payload


class APIResponse:

    def __init__(self, original_timestamp: float,
                 timestamp: float,
                 payload,
                 exception: Optional[Exception] = None,
                 ):
        self._original_timestamp = original_timestamp
        self._timestamp = timestamp
        self._payload = payload
        self._exception = exception

    @property
    def original_timestamp(self):
        return self._original_timestamp

    @property
    def timestamp(self):
        return self._timestamp

    @property
    def payload(self):
        return self._payload

    @property
    def exception(self):
        return self._exception



# Payload can come from the inbound process as empty as a request to be filled
# The inbound process is responsible for JSON conversion and validation
# The main process handles the actual data get/put - so no get/put/patch here

# class Payload:
#
#     _resource: Resource = None
#
#     def __init__(self, http_method: HTTPMethod, payload=None):
#         if not http_method.is_supported():
#             raise DE1ValueError(f"Unsupported HTTP method: {http_method}")
#         self._http_method = http_method
#         self._payload = payload
#         self._validated = False
#
#     # def get(self, dispatcher: Dispatcher):
#     #     raise NotImplementedError
#     #
#     # def put(self, dispatcher: Dispatcher):
#     #     raise NotImplementedError
#     #
#     # def patch(self, dispatcher: Dispatcher):
#     #     raise NotImplementedError
#
#     def as_json(self):
#         work = {k: fix_enums(v) for k, v in self.__dict__.items()
#                 if not k.startswith('_')}
#         return json.dumps(work)
#
#     @property
#     def payload(self):
#         return self._payload
#
#     @payload.setter
#     def payload(self, value):
#         raise NotImplementedError









"""
Copyright © 2021 Jeff Kletsky. All Rights Reserved.

License for this software, part of the pyDE1 package, is granted under
GNU General Public License v3.0 only
SPDX-License-Identifier: GPL-3.0-only
"""

import asyncio
import atexit
import enum
import logging
import re
import time
import traceback

from typing import Optional, Callable, Coroutine, Any, Union

from bleak import BleakScanner, BleakClient, BleakError
from bleak import BleakError, BleakClient, BleakScanner
from bleak.backends.bluezdbus.client import BleakClientBlueZDBus
from bleak.backends.device import BLEDevice
from bleak.backends.scanner import AdvertisementData

import pyDE1.default_logger
from pyDE1.bleak_client_wrapper import BleakClientWrapped
from pyDE1.de1.exceptions import DE1APIValueError
from pyDE1.dispatcher.resource import ConnectivityEnum
from pyDE1.event_manager import EventPayload, SubscribedEvent
from pyDE1.event_manager.events import ConnectivityState, ConnectivityChange
from pyDE1.scale.events import ScaleWeightUpdate, ScaleButtonPress, \
    ScaleTareSeen

logger = logging.getLogger('Scale')


class ScaleError(RuntimeError):
    def __init__(self, *args, **kwargs):
        super(ScaleError, self).__init__(args, kwargs)


class ScaleNoAddressError(ScaleError):
    def __init__(self, *args, **kwargs):
        super(ScaleNoAddressError, self).__init__(args, kwargs)


class ScaleNotConnectedError(ScaleError):
    def __init__(self, *args, **kwargs):
        super(ScaleNotConnectedError, self).__init__(args, kwargs)


class Scale:

    def __init__(self):
        self._address = None
        self._name: Optional[str] = None

        # These are often model-specific, override in subclass init
        self._nominal_period = 0.1  # seconds per sample
        self._minimum_tare_request_interval = 2.5 * self._nominal_period
        self._sensor_lag = 0.38  # seconds, including all delays to arrival
        # From https://www.youtube.com/watch?v=SIzFhnZ32Y0
        # (James Hoffmann) at 4:51
        #   Hiroia    0.20
        #   Skale     0.33
        #   Felicita  0.45
        #   Acaia     0.64
        self._tare_timeout = 1.0  # seconds until considered coincidence
        self._tare_threshold = 0.05  # grams, within this, considered "at zero"
        self.hold_at_tare = False

        self._bleak_client: Optional[BleakClientWrapped] = None

        self._event_connectivity: SubscribedEvent = SubscribedEvent(self)
        self._event_weight_update: SubscribedEvent = SubscribedEvent(self)
        self._event_button_press: SubscribedEvent = SubscribedEvent(self)
        self._event_tare_seen: SubscribedEvent = SubscribedEvent(self)
        self._estimated_period = self._nominal_period
        self._last_weight_update_received = 0
        self._last_tare_request_sent = 0

        # TODO: Think about how to manage "tare seen"
        #       Could use asyncio.Event(), but what are the states
        #       and how do you "release" if it never arrives?
        #       It seems like the use case would be:
        #           request tare
        #           wait for tare
        #           if seen:
        #               continue
        #           else:
        #               do something else
        self._tare_requested = False
        self._period_estimator = self.PeriodEstimator(self._nominal_period)

        # Don't need to await this on instantiation
        asyncio.get_event_loop().create_task(
            self._event_weight_update.subscribe(self._create_self_callback()))

    @classmethod
    def device_adv_is_recognized_by(cls, device: BLEDevice, adv: AdvertisementData):
        """
        Think about how to have this scan the subclasses
        for the most-specific match.

        Also think about how to manage the possibility of a collection
        of scales, that might already be instantiated
        """
        raise NotImplementedError

    @property
    def type(self):
        return self.__class__.__name__

    @property
    def address(self):
        addr = self._address
        if isinstance(addr, BLEDevice):
            addr = addr.address
        return addr

    @address.setter
    def address(self, address: Union[BLEDevice, str]):
        if self.address is not None:
            raise DE1APIValueError(
                "Changing the scale address is not yet supported")
        # If using BLEDevice directly
        # AttributeError: 'BleakClientBlueZDBus' object has no attribute 'lower'
        if isinstance(address, BleakClientBlueZDBus):
            self._address = address.address
        else:
            self._address = address

    @property
    def name(self):
        return self._name

    @property
    def sensor_lag(self):
        return self._sensor_lag

    async def connect(self, timeout=5.0):
        class_name = type(self).__name__
        logger.info(f"Connecting to {class_name} at {self.address}")
        if self._bleak_client is None:
            if self.address is None:
                raise ScaleNoAddressError
            self._bleak_client = BleakClientWrapped(self.address)
        self._bleak_client.set_disconnected_callback(
            self._create_disconnect_callback()
        )
        await asyncio.gather(self._event_connectivity.publish(
            ConnectivityChange(arrival_time=time.time(),
                               state=ConnectivityState.CONNECTING)),
            self._bleak_client.connect(timeout=timeout),
        )
        if self.is_connected:
            self._address = self._bleak_client.address
            if self.name is None:
                self._name = self._bleak_client.name
            logger.info(f"Connected to {class_name} at {self.address}")
            await self._event_connectivity.publish(
                ConnectivityChange(arrival_time=time.time(),
                                   state=ConnectivityState.CONNECTED))
        else:
            logger.error(
                f"Connection failed to {class_name} at {self.address}")
            await self._event_connectivity.publish(
                ConnectivityChange(arrival_time=time.time(),
                                   state=ConnectivityState.DISCONNECTED))

    async def standard_initialization(self, hold_notification=False):
        await self.display_on()
        await self.start_sending_weight_updates()
        if self.supports_button_press:
            await self.start_sending_button_updates()
        if not hold_notification:
            await self._notify_ready()

    async def _notify_ready(self):
        await self._event_connectivity.publish(
            ConnectivityChange(arrival_time=time.time(),
                               state=ConnectivityState.READY))

    async def _notify_not_ready(self):
        await self._event_connectivity.publish(
            ConnectivityChange(arrival_time=time.time(),
                               state=ConnectivityState.NOT_READY))

    async def disconnect(self):
        class_name = type(self).__name__
        logger.info(f"Disconnecting from {class_name}")
        if self._bleak_client is None:
            logger.info(f"Disconnecting from {class_name}; no client")
        await asyncio.gather(
            self._bleak_client.disconnect(),
            self._event_connectivity.publish(
                ConnectivityChange(arrival_time=time.time(),
                                   state=ConnectivityState.DISCONNECTING))
        )
        if self.is_connected:
            logger.error(
                f"Disconnect failed from {class_name} at {self.address}")
            await self._event_connectivity.publish(
                ConnectivityChange(arrival_time=time.time(),
                                   state=ConnectivityState.CONNECTED))
        else:
            logger.info(f"Scale.disconnect(): Disconnected from {class_name} at {self.address}")
            await self._event_connectivity.publish(
                ConnectivityChange(arrival_time=time.time(),
                                   state=ConnectivityState.DISCONNECTED))

    # TODO: Decide how to handle  self._disconnected_callback

    @property
    def is_connected(self):
        if self._bleak_client is None:
            return False
        else:
            return self._bleak_client.is_connected

    async def start_sending_weight_updates(self):
        raise NotImplementedError

    async def stop_sending_weight_updates(self):
        raise NotImplementedError

    @property
    def is_sending_weight_updates(self):
        raise NotImplementedError

    @property
    def supports_button_press(self):
        return False

    async def start_sending_button_updates(self):
        raise NotImplementedError

    async def stop_sending_button_updates(self):
        raise NotImplementedError

    async def tare(self):
        """
        A tare request can only be made every
        self._minimum_tare_request_interval seconds

        It doesn't make sense to hammer it as it will take
        at least one reporting period to "see" the tare
        """
        dt = time.time() - self._last_tare_request_sent
        if dt > self._minimum_tare_request_interval:
            await self._tare_internal()
            self._last_tare_request_sent = time.time()
            self._tare_requested = True
            logger.info(f"Tare request sent")
        else:
            logger.info(
                f"Tare request skipped, too soon, {dt:0.3f} seconds")
        return self._last_tare_request_sent


    async def _tare_internal(self):
        raise NotImplementedError

    async def current_weight(self):
        raise NotImplementedError

    async def display_on(self):
        raise NotImplementedError

    async def display_off(self):
        raise NotImplementedError

    # The two *_bool for API

    async def tare_with_bool(self, do_it=True):
        if do_it:
            await self.tare()

    async def display_bool(self, on):
        if on:
            await self.display_on()
        else:
            await self.display_off()

    @property
    def estimated_period(self):
        return self._estimated_period

    @property
    def event_weight_update(self):
        return self._event_weight_update

    @property
    def event_button_press(self):
        return self._event_button_press

    @property
    def event_tare_seen(self):
        return self._event_tare_seen

    def _scale_time_from_latest_arrival(self,
                                        latest_arrival: float):
        """
        Given the latest arrival, provide "best" estimate
        of when that weight was on the scale

        At present, just compensates for scale._scale_delay
        which should include transit delays and the like
        """
        return latest_arrival - self._sensor_lag

    def _update_scale_time_estimator(self,
                                     latest_arrival:float):
        """
        Call once per arrival to update any "fancy" algorithms such as PLL
        """
        pass

    def _create_self_callback(self) -> Coroutine:
        scale = self

        async def self_callback(swu: ScaleWeightUpdate) -> None:
            nonlocal scale
            dt = swu.arrival_time - scale._last_weight_update_received
            scale._last_weight_update_received = swu.arrival_time

            # TODO: Run profiler and evaluate if creating a task
            #       is consuming too much time

            asyncio.create_task(
                scale._period_estimator.process_arrival(dt))

            if scale._tare_requested:
                dt = swu.arrival_time - scale._last_tare_request_sent
                if dt > scale._tare_timeout:
                    scale._tare_requested = False
                    logger.error(f"No tare seen after {dt:0.03f} seconds")
                elif abs(swu.weight) < scale._tare_threshold:
                    scale._tare_requested = False
                    await scale.event_tare_seen.publish(
                        ScaleTareSeen(swu.arrival_time)
                    )
                    logger.info(f"Tare seen after {dt:0.03f} seconds")

            if scale.hold_at_tare:
                if abs(swu.weight) > scale._tare_threshold:
                    # Timing will be checked in scale.tare()
                    await scale.tare()

        return self_callback

    # TODO: Decide how to handle  self._disconnected_callback
    #   disconnected_callback (callable): Callback that will be scheduled in the
    #   event loop when the client is disconnected. The callable must take one
    #   argument, which will be this client object.

    # The callback seems to be expected to be a "plain" function
    # RuntimeWarning: coroutine 'DE1._create_disconnect_callback.<locals>.disconnect_callback'
    #                 was never awaited

    def _create_disconnect_callback(self) -> Callable:
        scale = self

        def disconnect_callback(client: BleakClientWrapped):
            nonlocal scale
            class_name = type(self).__name__
            logger.info(
                "disconnect_callback: "
                f"Disconnected from {class_name} at {client.address}, "
                "willful_disconnect: "
                f"{client.willful_disconnect}")
            asyncio.create_task(scale._event_connectivity.publish(
                ConnectivityChange(arrival_time=time.time(),
                                   state=ConnectivityState.DISCONNECTED)))

        return disconnect_callback

    @property
    def nominal_period(self):
        return self._nominal_period

    @nominal_period.setter
    def nominal_period(self, value):
        self._nominal_period = value
        self._period_estimator.reset(self._nominal_period)

    # Inner class
    class PeriodEstimator:
        """
        Estimate inter-arrival period from stream of arrivals

        Presently just an exponential moving average

        Skale II usually "bulks up" two or more reports on a 150-ms clock
        300 ms burbles aren't uncommon. A "normal" 50-ms stretch before its
        other half arrives would generate a (50/100) * k change.
        The other half then would generate (-100/100) * k change
        So k on the order of 1/1000 should be reasonable (10 sec, ~1 min settle)
        k of 1/10000 would be even better (100 sec, 10 min settle)
        Another way to look at this is 600 ms error / 600 s measurement ~ 0.1%

        Hand-in-hand with this is how long to consider a gap vs. a burble
        Nearly 5% of 150-ms windows from a SkaleII had 3 reports.
        Up to 6 in a window were observed. It dropped to 0.1% at 4 reports
        per window. Ignoring too many of these can lead to the estimate being off.
        Based on this, 300 ms (two periods) seems too short.
        300 + 150/2 = 375 ms is probably OK.
        450 + 150/2 = 525 ms is probablu conservative
        Try 500 ms to be reasonable.
        """

        def __init__(self, nominal_period: float):

            # TODO: How to update this for subclass changes?

            self._k = 1/10000
            self._ma = nominal_period
            self._too_long = 0.5  # seconds before considered a gap

            self._log_every_n = 1000
            self._n_counter = 0

        def reset(self, nominal_period: float):
            self._ma = nominal_period

        async def process_arrival(self, delta_arrival_time: float):

            if delta_arrival_time < self._too_long:
                self._ma = ((1 - self._k) * self._ma) \
                           + (self._k * delta_arrival_time)
                self._n_counter += 1
                if self._n_counter >= self._log_every_n:
                    self._n_counter = 0
                    logger.debug(f"Scale period: {self._ma}")


    # TODO: Deal with connectivity as a mixin

    # For API
    @property
    def connectivity(self):
        if self.is_connected:
            retval = ConnectivityEnum.CONNECTED
        else:
            # intentionally vague, as PUT "connecting" isn't needed
            retval = ConnectivityEnum.NOT_CONNECTED
        return retval

    async def connectivity_setter(self, value: ConnectivityEnum):
        assert isinstance(value, ConnectivityEnum), \
            f"mode of {value} not a ConnectivityEnum " \
            "in Scale.connectivity_setter()"
        if value is ConnectivityEnum.CONNECTED \
                and not self.is_connected:
            await self.connect()
        elif value is ConnectivityEnum.NOT_CONNECTED \
                and self.is_connected:
            await self.disconnect()


"""
Copyright © 2021 Jeff Kletsky. All Rights Reserved.

License for this software, part of the pyDE1 package, is granted under
GNU General Public License v3.0 only
SPDX-License-Identifier: GPL-3.0-only
"""

import asyncio
import enum
import logging
import sys
import time
from typing import Callable

from bleak.backends.device import BLEDevice
from bleak.backends.scanner import AdvertisementData

import pyDE1.default_logger
from pyDE1.scale import Scale
from pyDE1.scale.events import ScaleWeightUpdate, ScaleButtonPress

from pyDE1.event_manager.events import ConnectivityState, ConnectivityChange


logger = logging.getLogger('Scale.AtomaxIIScale')


class AtomaxSkaleII(Scale):

    def __init__(self):
        super(AtomaxSkaleII, self).__init__()
        self._nominal_period = 0.1  # seconds per sample
        self._minimum_tare_request_interval = 2.5 * self._nominal_period
        self._sensor_lag = 0.38  # seconds, including all transit delays
        self._tare_timeout = 1.0  # seconds until considered coincidence
        self._tare_threshold = 0.05  # grams, within this, considered "at zero"

        # Enable tare on button 1, hold UUID if need to unsubscribe
        self._button_1_tare_subscriber_id = None
        asyncio.get_event_loop().create_task(self._subscribe_button_press())

        # Linux, at least on an RPi 3B, needs response=True for write_gatt_char
        self._write_gatt_char_response = sys.platform == 'linux'

    class Char(enum.Enum):

        CONFIGURATION_EF80 =    'ef80'  # W
        WEIGHT_NOTIFY_EF81 =    'ef81'  # N
        BUTTON_NOTIFY_EF82 =    'ef82'  # N
        UNKNOWN_EF83 =          'ef83'  # R

        BATTERY_LEVEL =         '2a19'  # 0-63

        MODEL_NUMBER =          '2a24'  # "SkaleII"

        FW_REVISION =           '2a26'
        HW_REVISION =           '2a27'
        SW_REVISION =           '2a28'
        MANUFACTURER_NAME =     '2a29'  # "ATOMAX INC."

        UNKNOWN =               '0f050002-3225-44b1-b97d-d3274acb29de'  # R

        @property
        def cuuid(self):
            return f"0000{self.value}-0000-1000-8000-00805f9b34fb"


    # These typically get written to CONFIGURATION_EF80
    class Command(enum.Enum):

        DISPLAY_WEIGHT = b'\xec'
        DISPLAY_ON = b'\xed'
        DISPLAY_OFF = b'\xee'

        OUNCES = b'\x02'
        GRAMS = b'\x03'
        PERSIST_UNITS = b'\04'

        TARE = b'\x10'

        TIMER_ZERO = b'\xd0'
        TIMER_START = b'\xd1'
        TIMER_STOP = b'\xd2'

        # Place cal weight on Skale, send command
        CALIBRATE_500G = b'\xca\x05'
        CALIBRATE_1000G = b'\xca\x0a'

        # Undocumented at this time
        FILTER_WEAK = b'\x09\x00'
        FILTER_DEFAULT = b'\x09\x01'
        FILTER_STRONG = b'\x09\x02'
        FILTER_STRONGER = b'\x09\x03'

        @property
        def cuuid(self):
            return AtomaxSkaleII.Char.CONFIGURATION_EF80.cuuid

    @classmethod
    def device_adv_is_recognized_by(cls, device: BLEDevice,
                                    adv: AdvertisementData):
        if adv.local_name and adv.local_name.startswith('Skale'):
            return cls
        else:
            return None

    async def standard_initialization(self, hold_notification=False):
        await super(AtomaxSkaleII, self).standard_initialization(
            hold_notification=True)
        await self.set_grams()
        if not hold_notification:
            await self._event_connectivity.publish(
                ConnectivityChange(
                    arrival_time=time.time(),
                    state=ConnectivityState.READY
                )
            )

    async def update_self_from_device(self):
        self._model_number = await self._bleak_client.read_gatt_char(
            self.Char.MODEL_NUMBER.cuuid)
        self._fw_revision = await self._bleak_client.read_gatt_char(
            self.Char.FW_REVISION.cuuid)
        self._hw_revision = await self._bleak_client.read_gatt_char(
            self.Char.HW_REVISION.cuuid)
        self._sw_revision = await self._bleak_client.read_gatt_char(
            self.Char.SW_REVISION.cuuid)
        self._manufacturer_name = await self._bleak_client.read_gatt_char(
            self.Char.MANUFACTURER_NAME.cuuid)

    async def start_sending_weight_updates(self):
        await self._bleak_client.start_notify(
            self.Char.WEIGHT_NOTIFY_EF81.cuuid,
            self._create_weight_update_hander())
        logger.info("Sending weight updates")

    async def stop_sending_weight_updates(self):
        await self._bleak_client.stop_notify(
            self.Char.WEIGHT_NOTIFY_EF81.cuuid)
        logger.info("Stopped weight updates")

    def is_sending_weight_updates(self):
        return NotImplementedError

    async def _tare_internal(self):
        await self.send_command(self.Command.TARE)
        logger.info("Internal tare sent")

    async def current_weight(self):
        # await self._bleak_client.read_gatt_char(
        #     self.Char.UNKNOWN_EF83.uuid)
        # # That CUUID doesn't look like weight
        raise NotImplementedError

    async def display_on(self):
        await self.send_command(self.Command.DISPLAY_WEIGHT)
        await self.send_command(self.Command.DISPLAY_ON)
        logger.info("Display on")

    async def display_off(self):
        await self.send_command(self.Command.DISPLAY_OFF)
        logger.info("Display off")

    async def set_grams(self):
        await self.send_command(self.Command.GRAMS)
        logger.info("Grams selected")

    @property
    def supports_button_press(self):
        return True

    async def start_sending_button_updates(self):
        await self._bleak_client.start_notify(
            self.Char.BUTTON_NOTIFY_EF82.cuuid,
            self._create_button_press_hander())
        logger.info("Sending button updates")

    async def stop_sending_button_updates(self):
        await self._bleak_client.stop_notify(
            self.Char.BUTTON_NOTIFY_EF82.cuuid)
        logger.info("Stopped button updates")

    async def send_command(self, command: Command):
        await self._bleak_client.write_gatt_char(command.cuuid, command.value,
            response=self._write_gatt_char_response)

    def _create_weight_update_hander(self) -> Callable:
        scale = self

        async def weight_update_handler(sender, data):
            nonlocal scale

            now = time.time()

            w1 = int.from_bytes(data[1:4], byteorder='little',
                                signed=True) / 10.0
            w2 = int.from_bytes(data[5:8], byteorder='little',
                                signed=True) / 10.0
            # if w1 == w2:
            #     print(f"{dt:8.6f} {w1}")
            # else:
            #     print(f"{dt:8.6f} {w1}  {w2}")

            self._update_scale_time_estimator(now)

            await scale.event_weight_update.publish(
                ScaleWeightUpdate(
                    arrival_time=now,
                    scale_time=self._scale_time_from_latest_arrival(now),
                    weight=w1
                ))

        return weight_update_handler

    def _create_button_press_hander(self) -> Callable:
        scale = self

        async def button_press_handler(sender, data):
            nonlocal scale

            now = time.time()
            b = data[0]
            sbp = ScaleButtonPress(arrival_time=now, button=b)
            await scale.event_button_press.publish(sbp)

        return button_press_handler

    async def _subscribe_button_press(self):
        scale = self

        async def button_1_tare(sbp: ScaleButtonPress) -> None:
            nonlocal scale
            if sbp.button == 1:
                await scale.tare()
                logger.info("Button 1 - Tare requested")

        self._button_1_tare_subscriber_id \
            = await self._event_button_press.subscribe(button_1_tare)


"""
Copyright © 2021 Jeff Kletsky. All Rights Reserved.

License for this software, part of the pyDE1 package, is granted under
GNU General Public License v3.0 only
SPDX-License-Identifier: GPL-3.0-only


Utilities to be able to write a "shot file" containing the digested data
that describe how the shot progressed
"""
import logging
import time

from typing import Optional, Coroutine

import pyDE1.default_logger

from pyDE1.de1.c_api import API_MachineStates
from pyDE1.event_manager import SubscribedEvent
from pyDE1.de1.events import ShotSampleUpdate, ShotSampleWithVolumesUpdate
from pyDE1.scale.events import WeightAndFlowUpdate

# logger = logging.getLogger('ShotFile')

import aiologger
logger = aiologger.Logger.with_default_handlers()



async def basic_shot_sample_logger(ssu: ShotSampleWithVolumesUpdate):
    now = time.time()
    line = "{:5.2f} {:5.2f} {:6.1f} {:2d}     {:5d}  {:0.3f} {:0.3f} ms".format(
        ssu.group_pressure,
        ssu.group_flow,
        ssu.mix_temp,
        ssu.frame_number,
        ssu.sample_time,
        (ssu.create_time - ssu.arrival_time) * 1000,
        (now - ssu.arrival_time) * 1000,
    )
    logger.info(line)


async def gated_basic_shot_sample_logger(sswvu: ShotSampleWithVolumesUpdate):
    now = time.time()
    if sswvu.sender._recorder_active:
        line = "{:5.2f} {:5.2f} {:4.1f} {:2d} {:.1f} {:.1f} {:.1f} {} " \
                "{:5d}  {:0.3f} {:0.3f} ms".format(
            sswvu.group_pressure,
            sswvu.group_flow,
            sswvu.mix_temp,
            sswvu.frame_number,
            sswvu.volume_preinfuse,
            sswvu.volume_pour,
            sswvu.volume_total,
            [round(v,1) for v in sswvu.volume_by_frames],
            sswvu.sample_time,
            (sswvu.create_time - sswvu.arrival_time) * 1000,
            (now - sswvu.arrival_time) * 1000,
        )
        logger.info(line)


class CombinedShotLogger:

    def __init__(self):
        self._last_weight = 0
        self._last_flow = 0

        self._k_ma = 1/10
        self._show_ma_every = 100  # About 5/second

        self._create_ma_sswvu = 1
        self._end_to_end_ma_sswvu = 1
        self._show_ma_count_sswvu = 0

        self._create_ma_wafu = 1
        self._end_to_end_ma_wafu = 1
        self._show_ma_count_wafu = 0


    async def sswvu_subscriber(self, sswvu: ShotSampleWithVolumesUpdate):
        now = time.time()
        self._show_ma_count_sswvu += 1
        t_create = (sswvu.create_time - sswvu.arrival_time) * 1000
        t_end_to_end = (now - sswvu.arrival_time) * 1000
        self._create_ma_sswvu = self._create_ma_sswvu * (1 - self._k_ma) \
                                + t_create * self._k_ma
        self._end_to_end_ma_sswvu = self._end_to_end_ma_sswvu * (1 - self._k_ma) \
                                    + t_end_to_end * self._k_ma
        if self._show_ma_count_sswvu % self._show_ma_every == 0:
            logger.info(
                "SSWVU create, deliver, e2e: "
                "{:.3f}  {:.3f}  {:.3f} ms ".format(
                    self._create_ma_sswvu,
                    self._end_to_end_ma_sswvu - self._create_ma_sswvu,
                    self._end_to_end_ma_sswvu
                )
            )
        if sswvu.sender._recorder_active:
            line = "{:5.2f} b  {:4.2f} mL/s  {:4.2f} g/s  {:5.1f} g  " \
                   "{:4.1f} °C " \
                   "{:2d} {:.1f} {:.1f} {:.1f} {} " \
                   "{:5d}  {:0.3f} {:0.3f} ms".format(
                sswvu.group_pressure,
                sswvu.group_flow,
                self._last_flow,
                self._last_weight,
                sswvu.mix_temp,
                sswvu.frame_number,
                sswvu.volume_preinfuse,
                sswvu.volume_pour,
                sswvu.volume_total,
                [round(v, 1) for v in sswvu.volume_by_frames],
                sswvu.sample_time,
                t_create,
                t_end_to_end,
            )
            logger.info(line)

    async def wafu_subscriber(self, wafu: WeightAndFlowUpdate):
        now = time.time()
        self._show_ma_count_wafu += 1
        t_create = (wafu.create_time - wafu.arrival_time) * 1000
        t_end_to_end = (now - wafu.arrival_time) * 1000
        self._create_ma_wafu = self._create_ma_wafu * (1 - self._k_ma) \
                                + t_create * self._k_ma
        self._end_to_end_ma_wafu = self._end_to_end_ma_wafu * (1 - self._k_ma) \
                                    + t_end_to_end * self._k_ma
        if self._show_ma_count_wafu % self._show_ma_every == 0:
            logger.info(
                "WAFU create, deliver, e2e: "
                "{:.3f}  {:.3f}  {:.3f} ms ".format(
                    self._create_ma_wafu,
                    self._end_to_end_ma_wafu - self._create_ma_wafu,
                    self._end_to_end_ma_wafu
                )
            )
        self._last_flow = wafu.average_flow
        self._last_weight = wafu.current_weight



"""
Copyright © 2021 Jeff Kletsky. All Rights Reserved.

License for this software, part of the pyDE1 package, is granted under
GNU General Public License v3.0 only
SPDX-License-Identifier: GPL-3.0-only


Estimators to use with ScaleProcessor
"""

import logging

# from statistics import mean
from statistics import median
from typing import Tuple, List

import pyDE1.default_logger
from pyDE1.scale.processor import ScaleProcessor, ScaleProcessorError

logger = logging.getLogger('Estimator')

# Test to see how much of the ~4.5 ms delay is coming from calling mean()
# Quick check shows delay dropped to ~2.5 ms, so leaving False for now
# Resulting std. dev. is probably around 100 ms / sqrt(2)\

# TODO: Come back and revisit this decision
_USE_MEAN_FOR_TIME = True

# This drops the delay to ~2.5 ms as well
def mean(data: List[float]):
    return sum(data)/len(data)

# A perhaps good-enough median estimator (lower of even-length list)
# Still around 2.4 ms, so not really a significant gain -- use statistics.median
# def median(data: List[float]):
#     # .sort() is in-place
#     dcopy = data.copy()
#     dcopy.sort()
#     return dcopy[int(len(data)/2)]


class Estimator:
    """
    Generic estimator

    Writes the value into scale_processor.target_attr
    Writes the time into scale_processor.target_attr_time
    Assumes that scale_processor._history is a list

    In contrast to other implementations, the time estimates
    include scale_delay and estimated_period/2
    """

    def __init__(self, scale_processor: ScaleProcessor,
                 target_attr: str):
        self._scale_processor = scale_processor
        self._target_attr = target_attr
        # sanity check, don't be writing non-existent attributes
        # hasattr() "is implemented by calling getattr(object, name)"
        try:
            getattr(self._scale_processor, self._target_attr)
            getattr(self._scale_processor, self._target_attr + "_time")
        except AttributeError as e:
            raise e  # in case one day I get fancier, dev-only likely

        # Number of samples needed to estimate, set from subclass
        self._needed_internal: int = 1

    def estimate(self):
        if self._scale_processor._history_available >= self._needed:
            (val, tval) = self._estimate_inner()
            tval -= (self._scale_processor.scale.sensor_lag
                     + self._scale_processor.scale.nominal_period / 2)
        else:
            val = 0
            tval = 0
        setattr(self._scale_processor, self._target_attr, val)
        setattr(self._scale_processor, self._target_attr + "_time", tval)

    def _estimate_inner(self) -> Tuple[float, float]:
        raise NotImplementedError

    @property
    def _needed(self):
        return self._needed_internal

    @_needed.setter
    def _needed(self, needed: int):
        # Arguably this should grab _history_lock, but increases here
        # should be "safe" and recoverable. Keep as non-async
        self._needed_internal = needed
        sp = self._scale_processor
        sp._history_max = max(needed, sp._history_max)


class CurrentWeight (Estimator):

    def __init__(self, scale_processor: ScaleProcessor,
                 target_attr: str):
        super(CurrentWeight, self).__init__(scale_processor=scale_processor,
                                            target_attr=target_attr)
        self._needed = 1

    def _estimate_inner(self):
        val = self._scale_processor._history_weight[-1]
        tval = self._scale_processor._history_time[-1]
        return val, tval


class AverageFlow (Estimator):

    def __init__(self, scale_processor: ScaleProcessor,
                 target_attr: str,
                 samples: int):
        super(AverageFlow, self).__init__(scale_processor=scale_processor,
                                          target_attr=target_attr)
        self.samples = samples

    def _estimate_inner(self):
        # time data is jittery, use the best estimate
        dt = self.samples * self._scale_processor.scale.estimated_period
        val = ((self._scale_processor._history_weight[-1]
                - self._scale_processor._history_weight[-self.samples]) / dt)
        # (latest - dt/2) has a deviation of sigma + that of dt (small)
        # (latest + oldest)/2 has a deviation of sigma/sqrt(2) if independent
        # Following this, the average over the window should be even better
        if _USE_MEAN_FOR_TIME:
            tval = mean(self._scale_processor._history_time[-self.samples:])
        else:
            tval = (self._scale_processor._history_time[-self.samples]
                    + self._scale_processor._history_time[-1]) / 2
        return val, tval

    @property
    def samples(self):
        return self._samples

    @samples.setter
    def samples(self, value):
        self._samples = value
        self._needed = value


class MedianWeight (Estimator):

    def __init__(self, scale_processor: ScaleProcessor, target_attr: str,
                 samples: int):
        super(MedianWeight, self).__init__(scale_processor=scale_processor,
                                           target_attr=target_attr)
        self.samples = samples

    def _estimate_inner(self):
        val = median(self._scale_processor._history_weight[-self.samples:])
        if _USE_MEAN_FOR_TIME:
            tval = mean(self._scale_processor._history_time[-self.samples:])
        else:
            tval = (self._scale_processor._history_time[-self.samples]
                    + self._scale_processor._history_time[-1]) / 2
        return val, tval

    @property
    def samples(self):
        return self._samples

    @samples.setter
    def samples(self, value):
        self._samples = value
        self._needed = value


class MedianFlow (Estimator):
    """
    Estimate as finite difference between two medians

     samples_for_median                      samples_for_median
    -15 -14 -13 -12 -11 -10  -9  -8  -7  -6  -5  -4  -3  -2  -1
      |...............|                       |...............|
              |_______________________________________|
                               samples

    Similar to previous, deviation for average of the four points
    is better than that of one point offset by n * dt
    """

    def __init__(self, scale_processor: ScaleProcessor, target_attr: str,
                 samples: int, samples_for_medians: int):
        super(MedianFlow, self).__init__(scale_processor=scale_processor,
                                           target_attr=target_attr)
        self._samples = samples
        self._samples_for_medians = samples_for_medians
        # For side effects
        self.samples = samples
        self.samples_for_medians = samples_for_medians

    def _estimate_inner(self):
        p0 = -1
        p1 = -self.samples_for_medians
        p2 = -(1 + self.samples)
        p3 = -(self.samples + self.samples_for_medians)
        m0 = median(self._scale_processor._history_weight[p1:p0])
        m1 = median(self._scale_processor._history_weight[p3:p2])
        dt = self.samples * self._scale_processor.scale.estimated_period
        val = (m0 - m1)/dt
        if _USE_MEAN_FOR_TIME:
            tval = mean(self._scale_processor._history_time[p3:p0])
        else:
            tval = (self._scale_processor._history_time[p3]
                    + self._scale_processor._history_time[p0]) / 2
        return val, tval

    @property
    def samples(self):
        return self._samples

    @samples.setter
    def samples(self, value):
        self._samples = value
        self._needed = self.samples + self.samples_for_medians

    @property
    def samples_for_medians(self):
        return self._samples_for_medians

    @samples_for_medians.setter
    def samples_for_medians(self, value):
        self._samples_for_medians = value
        self._needed = self.samples + self.samples_for_medians


"""
Copyright © 2021 Jeff Kletsky. All Rights Reserved.

License for this software, part of the pyDE1 package, is granted under
GNU General Public License v3.0 only
SPDX-License-Identifier: GPL-3.0-only
"""

# TODO: Can't read a device if is is not connected
#       need a "resource temporarily unavailable" exception

# TODO: Wipe cached state on disconnect

import asyncio
import copy
import importlib.metadata as im
import inspect
import logging
import math  # for nan to be uniquely math.nan
import sys
import time

from typing import Optional, Union, Dict, Set

from pyDE1.de1.ble import CUUID
from pyDE1.de1.notifications import NotificationState
from pyDE1.utils import fix_enums
from pyDE1.utils_public import rgetattr, rsetattr

from pyDE1.dispatcher.resource import Resource
from pyDE1.dispatcher.mapping import MAPPING, IsAt

from pyDE1.de1 import DE1
from pyDE1.scale import Scale
from pyDE1.scale.processor import ScaleProcessor
from pyDE1.flow_sequencer import FlowSequencer
from pyDE1.de1.c_api import PackedAttr, ShotSettings, SetTime, \
    Versions, FWVersion, MMR0x80LowAddr, WaterLevels, WriteToMMR, \
    pack_one_mmr0x80_write, ReadFromMMR
from pyDE1.de1.exceptions import DE1APITypeError, DE1APIValueError, \
    DE1APIAttributeError, DE1APIKeyError
from pyDE1.dispatcher.mapping import IsAt

logger = logging.getLogger('APIImpl')

# If true, don't output nodes that have no value (write-only)
# or are empty dicts
# Otherwise math.nan fills in for the missing value
PRUNE_EMPTY_NODES = False


async def _value_to_property_setter(prop, value):

    if inspect.isroutine(prop):
        if inspect.iscoroutinefunction(prop):
            retval = await prop(value)
        else:
            retval = prop(value)
    else:
        raise DE1APITypeError (f"Setter {prop} is not an executable")

    return retval


# NB: This assumes that the MMR and CUUID are kept up to date
#     and that those that are read don't change on their own

async def _get_isat_value(isat: IsAt):

    # TODO: if IsAt.use_getter is enabled in the future, changes needed here

    target = isat.target
    attr_path = isat.attr_path
    if attr_path is None:
        raise DE1APIAttributeError(f"Write-only attribute {isat.__repr__()}")

    de1 = DE1()
    flow_sequencer = FlowSequencer()
    scale_processor = ScaleProcessor()
    scale = scale_processor.scale

    retval = None

    # For any attribute or property with a getter, getattr() "just works"
    # NB: property methords should not be coroutine functions

    if target is DE1:
        retval = rgetattr(de1, attr_path)

    elif target is FlowSequencer:
        retval = rgetattr(flow_sequencer, attr_path)

    elif target is Scale:
        retval = rgetattr(scale, attr_path)

    elif isinstance(target, MMR0x80LowAddr):
        # NB: This assumes that the MMR and CUUID are kept up to date
        #     and that those that are read don't change on their own

        if attr_path != '':
            raise DE1APIAttributeError(
                "MMR reads do not support attr_path "
                f"{isat}")

        # For now, assume everything is kept current

        # TODO: Unify with replication in PATCH operation

        try:
            retval = de1._mmr_dict[target].data_decoded
        except KeyError:
            retval = None
        if retval is None:
            t0 = time.time()
            ready = await de1.read_one_mmr0x80(target)
            await ready.wait()
            retval = de1._mmr_dict[target].data_decoded
            t1 = time.time()
            logger.debug(
                f"Read of {target.__repr__()} took \t"
                f"{(t1 - t0) * 1000:6.1f} ms"
            )

    elif inspect.isclass(target) and issubclass(target, PackedAttr):
        # NB: This assumes that the MMR and CUUID are kept up to date
        #     and that those that are read don't change on their own
        try:
            obj = (de1._cuuid_dict[target.cuuid]).last_value
        except KeyError:
            logger.info(
                f"No last value for {target.cuuid}"
            )
            obj = None
        if obj is None:
            t0 = time.time()
            obj = await de1.read_cuuid(target.cuuid)
            t1 = time.time()
            logger.debug(
                f"Read of {target} took \t{(t1 - t0) * 1000:6.1f} ms"
            )
        retval = rgetattr(obj, attr_path)

    else:
        raise DE1APITypeError(
            f"Mapping target of {target} is not recognized"
        )

    return fix_enums(retval)


async def _get_mapping_to_dict(partial_dict: dict) -> dict:
    """
    Takes a "branch" of a mapping and
      * Fills in any IsAt values
      * Recursively calls itself if a dict
      * Passes any other values unmodified
    """

    if isinstance(partial_dict, IsAt):
        partial_dict = { None: partial_dict }

    if not isinstance(partial_dict, dict):
        raise DE1APITypeError(f"Expected a dict, not {type(partial_dict)}")

    retval = {}

    for k, v in partial_dict.items():
        if isinstance(v, IsAt):
            try:
                this_val = await _get_isat_value(v)
            except AttributeError:
                if PRUNE_EMPTY_NODES:
                    continue # Don't write the key's entry
                else:
                    this_val = math.nan
        elif isinstance(v, dict):
            this_val = await _get_mapping_to_dict(v)
            # Suppress aggregates with nothing to aggregate
            if len(this_val) == 0 and PRUNE_EMPTY_NODES:
                continue
        else:
            this_val = v
        retval[k] = this_val

    return retval

# TODO: set and get versions

def _get_target_sets_inner(partial_dict: dict,
                           dict_of_sets: Dict[
                               str, Set[Union[PackedAttr, MMR0x80LowAddr]]],
                           include_can_read: bool,
                           include_can_write: bool):
    """
    Takes a "branch" of a mapping and returns a dict with two keys,
    'PacketAttr' and 'MMR0x80LowAddr', each with a set of targets.
    The dict_of_sets is modified in-place.
    """

    # See also validate.py
    # Valid: dict with dict
    #        IsAt with byte, bytearray (profile or firmware)
    if isinstance(partial_dict, IsAt):
        partial_dict = { None: partial_dict }

    if not isinstance(partial_dict, dict):
        raise DE1APITypeError(f"Expected a dict, not {type(partial_dict)}")

    for k, isat in partial_dict.items():

        if isinstance(isat, IsAt):
            target = isat.target
            if not ((inspect.isclass(target) and issubclass(target, PackedAttr))
                    or isinstance(target, MMR0x80LowAddr)):
                continue
            writable = target.can_write \
                       and not (isat.read_only
                                or (isat.attr_path is None
                                    and isat.setter_path is None))
            readable = target.can_read and isat.attr_path is not None
            if ((include_can_read and readable)
                    or (include_can_write and writable)):
                if isinstance(target, MMR0x80LowAddr):
                    dict_of_sets['MMR0x80LowAddr'].add(target)
                elif inspect.isclass(target) and issubclass(target, PackedAttr):
                    dict_of_sets['PackedAttr'].add(target)
            else:
                if not readable and not writable:
                    logger.error(
                        "IsAt should have either or both attr_path or setter_path "
                        f"not None: {isat.__repr__()}"
                    )


        elif isinstance(isat, dict):
            _get_target_sets_inner(isat, dict_of_sets,
                                   include_can_read, include_can_write)


def get_target_sets(mapping: dict,
                    include_can_read=False, include_can_write=False) \
        -> Dict[str, Set[Union[MMR0x80LowAddr, PackedAttr]]]:

    retval = {
        'MMR0x80LowAddr': set(),
        'PackedAttr': set()
    }

    _get_target_sets_inner(mapping, retval,
                           include_can_read, include_can_write)
    return retval


async def get_resource_to_dict(resource: Resource) -> dict:

    mapping = MAPPING[resource]
    de1 = DE1()
    flow_sequencer = FlowSequencer()

    # target_sets = get_target_sets(mapping)
    # pa_coros = list(map(lambda pa: de1.read_cuuid(pa.cuuid),
    #                     (target_sets['PackedAttr'])))
    # mmr_coros = list(map(lambda mmr: de1.read_one_mmr0x80(mmr),
    #                     (target_sets['MMR0x80LowAddr'])))
    # t0 = time.time()
    # # result = await asyncio.gather(*pa_coros) # 390-490 ms for 4
    # for coro in pa_coros:  # This takes 390 ms
    #     await coro
    # t1 = time.time()
    # logger.debug(
    #     f"Read of pa_coros took \t{(t1 - t0) * 1000:6.1f} ms"
    # )
    # t0 = time.time()
    # result = await asyncio.gather(*mmr_coros)  # Needs lock
    # await de1.read_standard_mmr_registers()  # This takes 1.2 seconds
    #                                          # and needs check for ready
    # t1 = time.time()
    # logger.debug(
    #     f"Read of mmr_coros took \t{(t1 - t0) * 1000:6.1f} ms"
    # )
    return await _get_mapping_to_dict(mapping)


# PATCH and PUT are related, but have slightly different requirements
#
# For PATCH, it is sufficient that each entry from the request has
# a corresponding entry in the overall mapping dictionary
#
# For PUT it is required that each element in the mappig dictionary
# have a corresponding entry in the request dictionary (completeness)
#
# TODO: How to handle read-only attributes in the PUT case is TBD

async def patch_resource_from_dict(resource: Resource, values_dict: dict):

    mapping = MAPPING[resource]

    # Get the list of PackedAttrs that need to be patched (properties and
    # MMRs are handled one at a time, as they are atomic), lock from changes,
    # get copies of current PackedAttrs to patch

    # TODO: Managing locks
    # TODO: PackedAttr "getter" that will wait on a pending update, if one is
    #       in flight, as well as unify the retrieval if not present

    # target_sets = get_target_sets(values_dict, include_can_write=True)

    # TODO: Fix this -- this needs a mapping, the way it is written now
    #       It really should only return those that are being changed
    #       in the case of a PATCH

    target_sets = get_target_sets(mapping, include_can_write=True)

    # Lock here
    de1 = DE1()
    flow_sequencer = FlowSequencer()

    for pa in target_sets['PackedAttr']:
        pa: PackedAttr
        cuuid = pa.cuuid
        try:
            last_value = de1._cuuid_dict[cuuid]._last_value
        except KeyError:
            last_value = None
        if last_value is None:
            t0 = time.time()
            last_value = await de1.read_cuuid(cuuid)
            t1 = time.time()
            logger.debug(
                f"Read of {cuuid} took \t"
                f"{(t1 - t0) * 1000:6.1f} ms"
            )

        # Don't load them all, just make sure they are there
        # Only load if they are being changed
        # pending_packed_attrs[pa] = last_value

    pending_packed_attrs = {}

    # Valid: dict with dict
    #        IsAt with byte, bytearray (profile or firmware)
    if isinstance(mapping, dict) \
            and isinstance(values_dict, dict):
        pass
    elif isinstance(mapping, IsAt) \
        and isinstance(values_dict, (bytes, bytearray)):
        # coerce into "standard form"
        mapping = { None: mapping }
        values_dict = { None: values_dict }
    else:
        raise DE1APITypeError(
            "Mapping and patch inconsistent, "
            "dict with dict, IsAt with raw value "
            f"not {type(mapping)} with {type(values_dict)}"
        )

    await _patch_dict_to_mapping_inner(values_dict, mapping,
                                       pending_packed_attrs)

    # if there are pending_packed_attrs, send them

    # Potentially gather, but may not be faster
    for pa in pending_packed_attrs.values():
        await de1.write_packed_attr(pa)

    # release locks

    return


async def _patch_dict_to_mapping_inner(partial_value_dict: dict,
                                       partial_mapping_dict: dict,
                                       pending_packed_attrs: Dict[
                                           type(PackedAttr), PackedAttr]):
    """
    This assumes that everything has been determined as "valid"
    """

    # TODO: PATCH Method for HTTP specifies that PATCH is to be atomic
    #       This isn't completely possible as there is no
    #       roll-back capability implemented at this time.
    #       To at least be somewhat compliant, all keys
    #       should be processed as fully as possible
    #       prior to writing any to their targets.
    #
    #       https://datatracker.ietf.org/doc/html/rfc5789
    #
    #       NB: Although this is similar to JSON Merge Patch
    #       https://datatracker.ietf.org/doc/html/rfc7386
    #       it is subtly different in that a JSON 'null'
    #       is interpreted as the valid, Python 'None'
    #       rather than as a request to delete an element:
    #
    #           This design means that merge patch documents are suitable for
    #           describing modifications to JSON documents that primarily
    #           use objects for their structure and do not make use of
    #           explicit null values.
    #
    #           The merge patch format is not appropriate for all JSON syntaxes.

    de1 = DE1()
    flow_sequencer = FlowSequencer()
    scale_processor = ScaleProcessor()
    scale = scale_processor.scale

    for key, new_value in partial_value_dict.items():

        try:
            mapping_isat = partial_mapping_dict[key]
        except KeyError:
            raise DE1APIKeyError(
                f"Unable to find mapping for {key} on the specified path"
            )

        if isinstance(mapping_isat, IsAt):
            target = mapping_isat.target
            attr_path = mapping_isat.attr_path
            setter_path = mapping_isat.setter_path

            if mapping_isat.read_only or attr_path is None and setter_path is None:
                raise DE1APIValueError(
                    f"Mapping for '{key}': {mapping_isat} is not writable"
                )

            if target in (DE1, FlowSequencer, Scale):
                if target is DE1:
                    this_target = de1
                elif target is FlowSequencer:
                    this_target = flow_sequencer
                elif target is Scale:
                    this_target = scale
                else:
                    raise DE1APITypeError(
                        f"Unsupported target for '{key}': {mapping_isat}")

                if setter_path is not None:
                    setter = rgetattr(this_target, setter_path)
                    await _value_to_property_setter(setter, new_value)
                else:
                    rsetattr(this_target, attr_path, new_value)

            elif isinstance(target, MMR0x80LowAddr):

                # MMR writes need to be serial and are atomic in that
                # each writable MMR is a single value.
                # As a result, just write it here and now.

                if attr_path != '':
                    raise DE1APIValueError(
                        "MMR writes do not support attr_path "
                        f"{target} {mapping_isat}")

                if setter_path is not None:
                    raise DE1APIValueError(
                        "MMR writes do not support setter_path "
                        f"{target} {mapping_isat}")

                mmr_write = pack_one_mmr0x80_write(
                    addr_low= target,
                    value= new_value,
                )

                logger.debug(f"MMR to be written: {mmr_write.as_wire_bytes()}")

                await de1.write_packed_attr(mmr_write)

                # TODO: Should this wait on ready.wait() ??
                #       Or is there a way to collect them all for later?
                ns: NotificationState = de1._mmr_dict[target]
                await ns.ready_event.wait()

            elif inspect.isclass(target) and issubclass(target, PackedAttr):
                # NB: This assumes that the CUUIDs are kept up to date
                #     and that those that are read don't change on their own

                packed_attr = (de1._cuuid_dict[target.cuuid]).last_value
                # TODO: Is this the right way to deal with these?
                if packed_attr is None:
                    packed_attr = await de1.read_cuuid(target.cuuid)
                old_value = rgetattr(packed_attr, attr_path)
                if new_value != old_value:
                    if not target in pending_packed_attrs:
                        pending_packed_attrs[target] = copy.deepcopy(packed_attr)
                    rsetattr(pending_packed_attrs[target], attr_path, new_value)

                # Send in outer method once all nodes are visited

            else:
                raise DE1APITypeError(
                    f"Mapping target of {target} is not recognized")

        elif isinstance(mapping_isat, dict):
            this_val = await _patch_dict_to_mapping_inner(
                partial_value_dict[key], partial_mapping_dict[key],
                pending_packed_attrs)
        else:
            this_val = new_value



"""
Copyright © 2021 Jeff Kletsky. All Rights Reserved.

License for this software, part of the pyDE1 package, is granted under
GNU General Public License v3.0 only
SPDX-License-Identifier: GPL-3.0-only
"""

import asyncio
import enum
import logging
import re

import pyDE1.default_logger
logger_cancel_tasks = logging.getLogger('CancelTasks')


def task_name_exists(name_to_find: str, starts_with=False):
    all_tasks = asyncio.all_tasks()
    found = False
    for t in all_tasks:
        name = t.get_name()
        if starts_with:
            found = name.startswith(name_to_find)
        else:
            found = name == name_to_find
        if found:
            break
    return found


def cancel_tasks_by_name(name_to_cancel: str, starts_with=False):
    all_tasks = asyncio.all_tasks()
    me = asyncio.current_task()
    for t in all_tasks:
        if t is me:
            continue
        name = t.get_name()
        if starts_with:
            cancel = name.startswith(name_to_cancel)
        else:
            cancel = name == name_to_cancel
        if cancel:
            logger_cancel_tasks.info(f"Canceling {t}")
            t.cancel()


def address_is_persistent(address: str) -> bool:
    """
    CoreBluetooth uses a UUID rather than a MAC address to identify devices
    As a result, it is (probably) not persistent over reboots.
    """
    if re.match(r'([0-9a-fA-F]{2}:){5}[0-9a-fA-F]{2}$', address):
        return True
    else:
        # macOS UUIDs don't seem to be persistent across Mac reboots
        return False


def fix_enums(val):
    """
    Return the name of an IntEnum, does not help with IntFlag
    So far no IntFlag enums headed to the external API
    """
    # Order is important due to IntEnum and IntFlag behavior (including int)
    if val is None or isinstance(val, (float, str, bool)):
        return val
    elif isinstance(val, enum.IntFlag):
        # Remove the class name
        return str(val).split('.',2)[1]
    elif isinstance(val, enum.IntEnum):
        return val.name
    elif isinstance(val, enum.Enum):
        return val.value
    elif isinstance(val, (bytearray, bytes)):
        return val.hex()
    else:
        return val



"""
Copyright © 2021 Jeff Kletsky. All Rights Reserved.

License for this software, part of the pyDE1 package, is granted under
GNU General Public License v3.0 only
SPDX-License-Identifier: GPL-3.0-only
"""

import asyncio
import logging
import time

from copy import deepcopy
from typing import Union, Optional

from pyDE1.de1.ble import CUUID
from pyDE1.de1.exceptions import DE1Error

from pyDE1.de1.c_api import PackedAttr, MMR0x80LowAddr, decode_one_mmr


class NotifyState ():

    def __init__(self, name: str):
        self.name = str(name)
        self.last_requested: Optional[float] = None
        self.last_updated: Optional[float] = None
        self._last_value: Optional[PackedAttr] = None
        self._ready_event: asyncio.Event = asyncio.Event()

    @property
    def last_value(self):
        # Paranoia, perhaps
        return deepcopy(self._last_value)

    @property
    def ready_event(self):
        return self._ready_event

    @property
    def update_complete(self):
        """
        last_updated is after last_requested
        :return: bool
        """
        return (self.last_requested is not None
                and self.last_updated is not None
                and self.last_updated > self.last_requested)

    def mark_requested(self, request_time=None):
        if request_time is None:
            request_time = time.time()
        self._ready_event.clear()
        self.last_requested = request_time

        return self._ready_event

    def mark_updated(self, obj: PackedAttr, update_time=None):
        if update_time is None:
            update_time = time.time()
        if self.last_requested is None:
            logger = logging.getLogger(f"{self.name}.Notify")
            logger.error(f"Update with no last_requested on {self.name}")
        self._last_value = obj
        self.last_updated = update_time
        self._ready_event.set()

        return self._ready_event


class NotificationState (NotifyState):

    def __init__(self, cuuid: CUUID):
        name = str(cuuid)
        super(NotificationState, self).__init__(name)
        self._cuuid = cuuid
        self._is_notifying = False

    @property
    def cuuid(self):
        return self._cuuid

    def mark_updated(self, obj: PackedAttr, update_time=None):
        super(NotificationState, self).mark_updated(obj, update_time)
        self._is_notifying = True

    @property
    def is_notifying(self):
        """
        There is a an acknowledged race condition here.
        It is possible for a stop_notifying to be requested
        and then a notification to be received.
        """
        return self._is_notifying

    def mark_ended(self):
        self._is_notifying = False



class MMR0x80Data (NotifyState):

    def __init__(self, addr_low: Union[MMR0x80LowAddr, int]):
        name = f"MMR0x80LowAddr.0x{addr_low:04x}"
        super(MMR0x80Data, self).__init__(name)
        self._addr_low = addr_low
        self._data_raw: Optional[Union[bytes, bytearray]] = None
        self._data_decoded: Optional[Union[bytes, bytearray,
                                           str, int, float, bool,
                                           dict]] = None

    @property
    def data_raw(self):
        return self._data_raw

    @data_raw.setter
    def data_raw(self, value):
        self._data_raw = value
        self._data_decoded = decode_one_mmr(0x80, self._addr_low, value)

    @property
    def data_decoded(self):
        return self._data_decoded





"""
Copyright © 2021 Jeff Kletsky. All Rights Reserved.

License for this software, part of the pyDE1 package, is granted under
GNU General Public License v3.0 only
SPDX-License-Identifier: GPL-3.0-only
"""
from http import HTTPStatus


class DE1Error (RuntimeError):
    def __init__(self, *args, **kwargs):
        super(DE1Error, self).__init__(args, kwargs)


class DE1ValueError (DE1Error, ValueError):
    def __init__(self, *args, **kwargs):
        super(DE1ValueError, self).__init__(args, kwargs)


class DE1TypeError (DE1Error, TypeError):
    def __init__(self, *args, **kwargs):
        super(DE1TypeError, self).__init__(args, kwargs)


class DE1AttributeError(DE1Error, AttributeError):
    def __init__(self, *args, **kwargs):
        super(DE1AttributeError, self).__init__(args, kwargs)


class DE1NoAddressError (DE1Error):
    def __init__(self, *args, **kwargs):
        super(DE1NoAddressError, self).__init__(args, kwargs)


class DE1NotConnectedError (DE1Error):
    def __init__(self, *args, **kwargs):
        super(DE1NotConnectedError, self).__init__(args, kwargs)


class DE1NoHandlerError(DE1Error):
    def __init__(self, *args, **kwargs):
        super(DE1NoHandlerError, self).__init__(args, kwargs)


class DE1ErrorStateReported(DE1Error):
    def __init__(self, *args, **kwargs):
        super(DE1ErrorStateReported, self).__init__(args, kwargs)


class DE1OperationInProgressError(DE1Error):
    def __init__(self, *args, **kwargs):
        super(DE1OperationInProgressError, self).__init__(args, kwargs)


class DE1APIError (DE1Error):
    def __init__(self, *args, **kwargs):
        super(DE1APIError, self).__init__(*args, **kwargs)


class DE1APITypeError (TypeError, DE1APIError):
    def __init__(self, *args, **kwargs):
        super(DE1APITypeError, self).__init__(*args, **kwargs)


class DE1APIValueError (ValueError, DE1APIError):
    def __init__(self, *args, **kwargs):
        super(DE1APIValueError, self).__init__(*args, **kwargs)


class DE1APIUnsupportedStateTransitionError (DE1APIValueError):
    def __init__(self, target_state, current_state, *args, **kwargs):
        message = "I'm afraid I can't do that Dave." \
                  f" Can't move to {target_state} from {current_state}"
        super(DE1APIUnsupportedStateTransitionError, self).__init__(
            message, *args, **kwargs)


class DE1APIAttributeError (AttributeError, DE1APIError):
    def __init__(self, *args, **kwargs):
        super(DE1APIAttributeError, self).__init__(*args, **kwargs)


class DE1APIKeyError(AttributeError, DE1APIError):
    def __init__(self, *args, **kwargs):
        super(DE1APIKeyError, self).__init__(*args, **kwargs)


class DE1APITooManyFramesError (DE1APIValueError):
    def __init__(self, *args, **kwargs):
        super(DE1APITooManyFramesError, self).__init__(*args, **kwargs)


class MMRTypeError (DE1APITypeError):
    def __init__(self, *args, **kwargs):
        super(MMRTypeError, self).__init__(*args, **kwargs)


class MMRValueError (DE1APIValueError):
    def __init__(self, *args, **kwargs):
        super(MMRValueError, self).__init__(*args, **kwargs)


class MMRDataTooLongError (MMRValueError):
    def __init__(self, *args, **kwargs):
        super(MMRDataTooLongError, self).__init__(*args, **kwargs)


class MMRAddressError (MMRValueError):
    def __init__(self, *args, **kwargs):
        super(MMRAddressError, self).__init__(*args, **kwargs)


class MMRAddressRangeError (MMRAddressError):
    def __init__(self, *args, **kwargs):
        super(MMRAddressRangeError, self).__init__(*args, **kwargs)


class MMRAddressOffsetError (MMRAddressError):
    def __init__(self, *args, **kwargs):
        super(MMRAddressOffsetError, self).__init__(*args, **kwargs)


"""
Copyright © 2021 Jeff Kletsky. All Rights Reserved.

License for this software, part of the pyDE1 package, is granted under
GNU General Public License v3.0 only
SPDX-License-Identifier: GPL-3.0-only

"Null" outbound processor that just keeps the queue clear,
counting the various messages by type,
logging every update_period seconds (and resetting the count)
"""

# Only import the minimal here, as it potentially ends up in all processes.\

import multiprocessing, multiprocessing.connection
import logging

# TODO: look into how loggers here relate to the root logger from "main"

# TODO: Look into or resolve processes' loggers writing over each other
import time


def run_api_outbound(api_outbound_queue: multiprocessing.Queue):
    logger = logging.getLogger('outbound')

    import asyncio
    import json

    async def run(api_outbound_queue: multiprocessing.Queue):

        last_update = time.time()
        update_period = 10  # in seconds
        counts = {}

        while True:
            item = api_outbound_queue.get()
            as_dict = json.loads(item)
            now = time.time()
            try:
                counts[as_dict['class']] += 1
            except KeyError:
                counts[as_dict['class']] = 1
            if now - last_update > update_period:
                logger.info(counts)
                counts = {}
                last_update = now

    loop = asyncio.get_event_loop()
    loop.set_debug(True)
    loop.run_until_complete(run(api_outbound_queue=api_outbound_queue))
